import signal
import sys
import time

import os

from artemis.config import get_artemis_config_value
from artemis.remote.child_processes import PythonChildProcess, RemotePythonProcess
from artemis.remote.remote_execution import ParamikoPrintThread, execute_command
from artemis.remote.utils import get_local_ips, get_remote_artemis_path
from functools import partial

ip_addresses = [get_artemis_config_value(section="tests", option="remote_server",default_generator=lambda: "127.0.0.1")]
# ip_addresses=["127.0.0.1"]
if ip_addresses[0] not in get_local_ips():
    ip_addresses.append("127.0.0.1")

def get_test_functions_path(ip_address):
    if ip_address in get_local_ips():
        return os.path.join(os.path.dirname(__file__), "remote_test_functions.py")
    else:
        remote_path = get_remote_artemis_path(ip_address)
        return os.path.join(remote_path, "remote/remote_test_functions.py")


def test_simple_pcp():
    for ip_address in ip_addresses:
        command = "python %s --callback=%s"%(get_test_functions_path(ip_address),"success_function")
        pyc = PythonChildProcess(ip_address=ip_address,command=command)
        (stdin, stdout, stderr) = pyc.execute_child_process()
        stderr_out =stderr.readlines()
        stderr_out = [line.strip() for line in stderr_out if "pydev debugger" not in line]
        stdout_out = stdout.readlines()
        assert len("".join(stderr_out)) == 0, "Stderr not empty. Received: %s"%stderr_out
        assert len(stdout_out) == 1 and stdout_out[0].strip() == "Success", "Stdout not as expected. Received: %s"%stdout_out

def test_simple_pcp_list():
    for ip_address in ip_addresses:
        command = ["python", get_test_functions_path(ip_address), "--callback=success_function"]
        pyc = PythonChildProcess(ip_address=ip_address,command=command)
        (stdin, stdout, stderr) = pyc.execute_child_process()
        stderr_out = stderr.readlines()
        stderr_out = [line.strip() for line in stderr_out if "pydev debugger" not in line]
        stdout_out = stdout.readlines()
        assert len("".join(stderr_out)) == 0, "Stderr not empty. Received: %s"%stderr_out
        assert len(stdout_out) == 1 and stdout_out[0].strip() == "Success", "Stdout not as expected. Received: %s"%stdout_out

def test_interrupt_process_gently():
    for ip_address in ip_addresses:
        command = ["python", get_test_functions_path(ip_address), "--callback=count_high"]

        cp = PythonChildProcess(ip_address, command)
        stdin , stdout, stderr = cp.execute_child_process()
        time.sleep(5)
        cp.kill()
        time.sleep(1)
        stdout_out = stdout.readlines()
        stderr_out = stderr.readlines()
        if cp.is_local():
            assert stderr_out[-1].strip() == "KeyboardInterrupt"
        else:
            assert stdout_out[-1].strip() == "KeyboardInterrupt"
        assert not cp.is_alive(), "Process is still alive, killing it did not work"

def test_kill_process_gently():
    for ip_address in ip_addresses:
        command = ["python", get_test_functions_path(ip_address), "--callback=sleep_function"]

        cp = PythonChildProcess(ip_address, command)
        stdin , stdout, stderr = cp.execute_child_process()
        time.sleep(1)
        cp.kill()
        time.sleep(1)
        stdout_out = stdout.readlines()
        assert stdout_out[-1].strip() == "Interrupted"
        assert not cp.is_alive(), "Process is still alive, killing it did not work"

def test_kill_process_strongly():
    for ip_address in ip_addresses:
        command = ["python", get_test_functions_path(ip_address), "--callback=hanging_sleep_function"]

        cp = PythonChildProcess(ip_address, command)
        stdin , stdout, stderr = cp.execute_child_process()
        time.sleep(1)
        cp.kill()
        time.sleep(1)
        assert cp.is_alive(), "Process terminated too soon. Check remote_test_functions.py implementation!"
        cp.kill(signal.SIGKILL)
        time.sleep(1)
        assert not cp.is_alive(), "Process is still alive, killing it did not work"



def test_remote_graphics():
    for ip_address in ip_addresses:
        command = ["python", get_test_functions_path(ip_address), "--callback=remote_graphics"]

        cp = PythonChildProcess(ip_address=ip_address,command=command,take_care_of_deconstruct=True)
        i, stdout, stderr = cp.execute_child_process()
        time.sleep(1)
        stderr_out = stderr.readlines()
        stdout_out = stdout.readlines()
        # assert len(stderr_out) == 0, "Stderr not empty. Received: %s"%stderr_out
        assert stdout_out[-1].strip() == "Success", "Stdout not as expected. Received: %s"%stdout_out

        # Do not call decons
        # cp.deconstruct(signum=signal.SIGINT)

def remote_test_func(a, b):
    print('a+b={}'.format(a+b))
    return a+b


def test_remote_child_function():

    for ip_address in ip_addresses:
        pyc = PythonChildProcess(ip_address=ip_address,command=partial(remote_test_func, a=1, b=2))
        (stdin, stdout, stderr) = pyc.execute_child_process()
        stderr_out =stderr.readlines()
        stderr_out = [line.strip() for line in stderr_out if "pydev debugger" not in line]
        stdout_out = stdout.readlines()


def my_func(a, b):
    print('hello hello hello')
    time.sleep(0.01)
    return a+b


def test_remote_python_process():

    in_debug_mode = sys.gettrace() is not None

    p = RemotePythonProcess(
        function=partial(my_func, a=1, b=2),
        ip_address='localhost',
        )

    stdin , stdout, stderr = p.execute_child_process()
    time.sleep(.1)  # That autta be enough

    errtext = stderr.read()
    if in_debug_mode:
        assert errtext.startswith('pydev debugger: ')
    else:
        assert errtext == '', errtext
    assert stdout.read() == 'hello hello hello\n'
    assert p.get_return_value()==3


if __name__ == "__main__":
    test_simple_pcp()
    test_simple_pcp_list()
    test_interrupt_process_gently()
    test_kill_process_gently()
    test_kill_process_strongly()
    test_remote_graphics()
    test_remote_child_function()
    test_remote_python_process()
    print("Tests finished")



import time
from collections import OrderedDict

import numpy as np

from artemis.general.checkpoint_counter import CheckPointCounter
from artemis.general.should_be_builtins import bad_value
from artemis.ml.tools.costs import get_evaluation_function
from artemis.ml.tools.iteration import checkpoint_minibatch_index_generator
from artemis.general.mymath import sqrtspace
from artemis.ml.tools.processors import RunningAverage


def compare_predictors(dataset, online_predictors={}, offline_predictors={}, minibatch_size = 'full',
        evaluation_function = 'mse', test_epochs = sqrtspace(0, 1, 10), report_test_scores = True,
        test_on = 'training+test', test_batch_size = None, accumulators = None, online_test_callbacks = {}):
    """
    DEPRECATED: use train_and_test_online_predictor instead.

    Compare a set of predictors by running them on a dataset, and return the learning curves for each predictor.

    :param dataset: A DataSet object
    :param online_predictors: A dict<str:IPredictor> of online predictors.  An online predictor is
        sequentially fed minibatches of data and updates its parameters with each minibatch.
    :param offline_predictors: A dict<str:object> of offline predictors.  Offline predictors obey sklearn's
        Estimator/Predictor interfaces - ie they methods
            estimator = object.fit(data, targets) and
            prediction = object.predict(data)
    :param minibatch_size: Size of the minibatches to use for online predictors.  Can be:
        An int, in which case it represents the minibatch size for all classifiers.
        A dict<str: int>, in which case you can set the minibatch size per-classifier.
        In place of the int, you can put 'all' if you want to train on the whole dataset in each iteration.
    :param test_epochs: Test points to use for online predictors.  Can be:
        A list of integers - in which case the classifier is after seeing this many samples.
        A list of floats - in which case the classifier is tested after seeing this many epochs.
        'always' - In which case a test is performed after every training step
        The final test point determines the end of training.
    :param evaluation_function: Function used to evaluate output of predictors
    :param report_test_scores: Boolean indicating whether you'd like to report results online.
    :param test_on: 'training', 'test', 'training+test'
    :param test_batch_size: When the test set is too large to process in one step, use this to break it
        up into chunks.
    :param accumulators: A dict<str: accum_fcn>, where accum_fcn is a stateful-function of the form:
        accmulated_output = accum_fcn(this_output)
        Special case: accum_fcn can be 'avg' to make a running average.
    :param online_test_callbacks: A dict<str: fcn> where fcn is a callback that takes an online
        predictor as an argument.  Useful for logging/plotting/debugging progress during training.
    :return: An OrderedDict<LearningCurveData>
    """

    all_keys = online_predictors.keys()+offline_predictors.keys()
    assert len(all_keys) > 0, 'You have to give at least one predictor.  Is that too much to ask?'
    assert len(all_keys) == len(np.unique(all_keys)), "You have multiple predictors using the same names. Change that."
    type_constructor_dict = OrderedDict(
        [(k, ('offline', offline_predictors[k])) for k in sorted(offline_predictors.keys())] +
        [(k, ('online', online_predictors[k])) for k in sorted(online_predictors.keys())]
        )

    minibatch_size = _pack_into_dict(minibatch_size, expected_keys=online_predictors.keys())
    accumulators = _pack_into_dict(accumulators, expected_keys=online_predictors.keys())
    online_test_callbacks = _pack_into_dict(online_test_callbacks, expected_keys=online_predictors.keys(), allow_subset=True)
    test_epochs = np.array(test_epochs)
    if isinstance(evaluation_function, str):
        evaluation_function = get_evaluation_function(evaluation_function)

    records = OrderedDict()

    # Run the offline predictors
    for predictor_name, (predictor_type, predictor) in type_constructor_dict.items():
        print('%s\nRunning predictor %s\n%s' % ('='*20, predictor_name, '-'*20))
        records[predictor_name] = \
            assess_offline_predictor(
                predictor=predictor,
                dataset = dataset,
                evaluation_function = evaluation_function,
                report_test_scores = report_test_scores,
                test_on = test_on,
                test_batch_size = test_batch_size
                ) if predictor_type == 'offline' else \
            assess_online_predictor(
                predictor=predictor,
                dataset = dataset,
                evaluation_function = evaluation_function,
                test_epochs = test_epochs,
                accumulator = accumulators[predictor_name],
                minibatch_size = minibatch_size[predictor_name],
                report_test_scores = report_test_scores,
                test_on = test_on,
                test_batch_size = test_batch_size,
                test_callback=online_test_callbacks[predictor_name] if predictor_name in online_test_callbacks else None
                ) if predictor_type == 'online' else \
            bad_value(predictor_type)

    print('Done!')

    return records


def _pack_into_dict(value_or_dict, expected_keys, allow_subset = False):
    """
    Used for when you want to either
        a) Distribute some value to all predictors
        b) Distribute different values to different predictors and check that the names match up.
    :param value_or_dict: Either
        a) A value
        b) A dict<predictor_name: value_for_predictor>
    :param expected_keys: Names of predictors
    :return: A dict<predictor_name: value_for_predictor>
    """
    if not isinstance(value_or_dict, dict):
        output_dict = {predictor_name: value_or_dict for predictor_name in expected_keys}
    else:
        output_dict = value_or_dict
        if allow_subset:
            assert set(value_or_dict.keys()).issubset(expected_keys), 'Expected a subset of: %s.  Got %s' % (expected_keys, value_or_dict.keys())
        else:
            assert set(expected_keys) == set(value_or_dict.keys()), 'Expected keys: %s.  Got %s' % (expected_keys, value_or_dict.keys())
    return output_dict


def dataset_to_testing_sets(dataset, test_on = 'training+test'):
    return \
        {'Training': (dataset.training_set.input, dataset.training_set.target), 'Test': (dataset.test_set.input, dataset.test_set.target)} if test_on == 'training+test' else \
        {'Test': (dataset.test_set.input, dataset.test_set.target)} if test_on == 'test' else \
        {'Training': (dataset.training_set.input, dataset.training_set.target)} if test_on == 'training' else \
        bad_value(test_on)


def assess_offline_predictor(predictor, dataset, evaluation_function, test_on = 'training+test', report_test_scores=True,
        test_batch_size = None):
    """
    Train an offline predictor and return the LearningCurveData (which will not really be a curve,
    but just a point representing the final score.

    :param predictor:  An object with methods fit(X, Y), predict(X)
    :param dataset: A DataSet object
    :param evaluation_function: A function of the form: score=fcn(actual_values, target_values)
    :param report_test_scores: Print out the test scores as they're computed (T/F)
    :return: LearningCurveData containing the score on the test sets
    """
    record = LearningCurveData()
    predictor.fit(dataset.training_set.input, dataset.training_set.target)
    testing_sets = dataset_to_testing_sets(dataset, test_on)
    scores = [(k, evaluation_function(process_in_batches(predictor.predict, x, test_batch_size), y)) for k, (x, y) in testing_sets.items()]
    record.add(None, scores)
    if report_test_scores:
        print('Scores: %s' % (scores, ))
    return record


def assess_online_predictor(predictor, dataset, evaluation_function, test_epochs, minibatch_size, test_on = 'training+test',
        accumulator = None, report_test_scores=True, test_batch_size = None, test_callback = None):
    """
    DEPRECATED: use assess_prediction_functions_on_generator in train_and_test_old.py

    Train an online predictor and return the LearningCurveData.

    :param predictor:  An IPredictor object
    :param dataset: A DataSet object
    :param evaluation_function: A function of the form: score=fcn(actual_values, target_values)
    :param test_epochs: List of epochs to test at.  Eg. [0.5, 1, 2, 4]
    :param minibatch_size: Number of samples per minibatch, or:
        'full' to do full-batch.
        'stretch': to stretch the size of each batch so that we make just one call to "train" between each test.  Use
            this, for instance, if your predictor trains on one sample at a time in sequence anyway.
    :param report_test_scores: Print out the test scores as they're computed (T/F)
    :param test_callback: A callback which takes the predictor, and is called every time a test
        is done.  This can be useful for plotting/debugging the state.
    :return: LearningCurveData containing the score on the test sets

    """
    # TODO: Remove this class, as it is deprecated

    record = LearningCurveData()

    testing_sets = dataset_to_testing_sets(dataset, test_on)
    if accumulator is None:
        prediction_functions = {k: predictor.predict for k in testing_sets}
    else:
        accum_constructor = {'avg': RunningAverage}[accumulator] \
            if isinstance(accumulator, str) else accumulator
        accumulators = {k: accum_constructor() for k in testing_sets}
        prediction_functions = {k: lambda inp, kp=k: accumulators[kp](predictor.predict(inp)) for k in testing_sets}
        # Bewate the in-loop lambda - but I think we're ok here.

    if isinstance(evaluation_function, str):
        evaluation_function = get_evaluation_function(evaluation_function)

    def do_test(current_epoch):
        scores = [(k, evaluation_function(process_in_batches(prediction_functions[k], x, test_batch_size), y)) for k, (x, y) in testing_sets.items()]
        if report_test_scores:
            print('Scores at Epoch %s: %s, after %.2fs' % (current_epoch, ', '.join('%s: %.3f' % (set_name, score) for set_name, score in scores), time.time()-start_time))
        record.add(current_epoch, scores)
        if test_callback is not None:
            record.add(current_epoch, ('callback', test_callback(predictor)))

    start_time = time.time()
    if minibatch_size == 'stretch':
        test_samples = (np.array(test_epochs) * dataset.training_set.n_samples).astype(int)
        i=0
        if test_samples[0] == 0:
            do_test(i)
            i += 1
        for indices in checkpoint_minibatch_index_generator(n_samples=dataset.training_set.n_samples, checkpoints=test_samples, slice_when_possible=True):
            predictor.train(dataset.training_set.input[indices], dataset.training_set.target[indices])
            do_test(test_epochs[i])
            i += 1
    else:
        checker = CheckPointCounter(test_epochs)
        last_n_samples_seen = 0
        for (n_samples_seen, input_minibatch, target_minibatch) in \
                dataset.training_set.minibatch_iterator(minibatch_size = minibatch_size, epochs = float('inf'), single_channel = True):
            current_epoch = (float(last_n_samples_seen))/dataset.training_set.n_samples
            last_n_samples_seen = n_samples_seen
            time_for_a_test, done = checker.check(current_epoch)
            if time_for_a_test:
                do_test(current_epoch)
            if done:
                break
            predictor.train(input_minibatch, target_minibatch)

    return record


def process_in_batches(func, data, batch_size):
    """
    Sometimes a function requires too much internal memory, so you have to process things in batches.
    """
    if batch_size is None:
        return func(data)

    n_samples = len(data)
    chunks = np.arange(int(np.ceil(float(n_samples)/batch_size))+1)*batch_size
    assert len(chunks)>1
    out = None
    for ix_start, ix_end in zip(chunks[:-1], chunks[1:]):
        x = data[ix_start:ix_end]
        y = func(x)
        if out is None:
            out = np.empty((n_samples, )+y.shape[1:], dtype = y.dtype)
            out[ix_start:ix_end] = y
    return out


class LearningCurveData(object):
    """
    A container for the learning curves resulting from running a predictor
    on a dataset.  Use this object to incrementally write results, and then
    retrieve them as a whole.
    """
    def __init__(self):
        self._times = {}
        self._scores = OrderedDict()
        self._latest_score = None

    def add(self, time, scores):
        """
        :param time: Something representing the time at which the record was taken.
        :param scores: A list of 2-tuples of (score_name, score).  It can also be a scalar score.
            Eg: [('training', 0.104), ('test', 0.119)]
        :return:
        """
        if np.isscalar(scores):
            scores = [('Score', scores)]
        elif isinstance(scores, tuple):
            scores = [scores]
        else:
            assert isinstance(scores, list) and all(len(s) == 2 for s in scores)

        for k, v in scores:
            if k not in self._scores:
                self._times[k] = []
                self._scores[k] = []
            self._times[k].append(time)
            self._scores[k].append(v)

    def get_results(self):
        """
        :return: (times, results), where:
            times is a length-N vector indicating the time of each test
            scores is a (length_N, n_scores) array indicating the each score at each time
                OR a (length_N, n_scores, n_reps) array where n_reps indexes each repetition or the same experiment
        """
        return {k: np.array(t) for k, t in self._times.items()}, OrderedDict((k, np.array(v)) for k, v in self._scores.items())

    def get_scores(self, which_test_set = None):
        """
        :return: scores for the given test set.
            For an offline predictor, scores'll be float
            For an online predictor, scores'll by a 1-D array representing the score at each test point.
        """
        _, results = self.get_results()

        if which_test_set is None:
            assert len(results)==1, 'You failed to specify which test set to use, which would be fine if there was only ' \
                "one, but there's more than one.  There's %s" % (results.keys(), )
            return next(v for v in results.values())
        else:
            assert which_test_set in results, 'You asked for results for the test set %s, but we only have test sets %s' \
                % (which_test_set, results.keys())
            return results[which_test_set]



import numpy as np
from artemis.experiments.decorators import experiment_function
from matplotlib import pyplot as plt
from six.moves import xrange

__author__ = 'peter'


"""
This file demonstates Artemis's "Experiments"

When you run an experiment, all figures and console output, as well as some metadata such as total run time, arguments,
etc are saved to disk.

This demo illustrates how you can create an experiment, create variations on that experiment, and view the results.
"""


class OnlineLinearRegressor:

    def __init__(self, n_in, n_out, learning_rate = 0.01):
        self.w = np.zeros((n_in, n_out))
        self.learning_rate = learning_rate

    def train(self, x, targ):  # x: (n_samples, n_in), targ: (n_samples, n_out)
        y = self.predict(x)
        self.w -= self.learning_rate * (x.T.dot(y-targ))

    def predict(self, x):  # x: (n_samples, n_in)
        return x.dot(self.w)


@experiment_function
def demo_linear_regression(
        n_in = 100,
        n_out = 4,
        n_training_samples = 500,
        n_test_samples = 500,
        noise = .1,
        n_epochs = 10,
        eta = 0.001,
        random_seed = 1234,
        score_report_period = 100,
        ):
    """
    Generate a random linear regression problem and train an online predictor to solve it with Stochastic gradient descent.
    Log the scores and plot the resulting learning curves.

    :param n_in: Number of inputs
    :param n_out: Number of outputs
    :param n_training_samples: Number of training samples in generated dataset.
    :param n_test_samples: Number of test samples in generated dataset.
    :param noise: Noise to add to generated dataset
    :param n_epochs: Number of epochs to run for
    :param eta: Learning rate for SGD
    :param random_seed: Random seed (for generating data)
    :param score_report_period: Report score every X training iterations.
    """

    # Setup data
    rng = np.random.RandomState(random_seed)
    w_true = rng.randn(n_in, n_out)*.1  # (n_in, n_out)
    training_data = rng.randn(n_training_samples, n_in)  # (n_training_samples, n_in)
    training_target = training_data.dot(w_true) + noise*rng.randn(n_training_samples, n_out)  # (n_training_samples, n_out)
    test_data = rng.randn(n_test_samples, n_in)  # (n_test_samples, n_in)
    test_target = test_data.dot(w_true) + noise*rng.randn(n_test_samples, n_out)  # (n_test_samples, n_out)
    predictor = OnlineLinearRegressor(n_in=n_in, n_out=n_out, learning_rate=eta)

    # Train and periodically record scores.
    epoch_scores = []
    for i in xrange(n_training_samples*n_epochs+1):
        if i % score_report_period == 0:
            training_out = predictor.predict(training_data)
            training_cost = ((training_target-training_out)**2).sum(axis=1).mean(axis=0)
            test_out = predictor.predict(test_data)
            test_cost = ((test_target-test_out)**2).sum(axis=1).mean(axis=0)
            print('Epoch {epoch}: Test Cost: {test}, Training Cost: {train}'.format(epoch=float(i)/n_training_samples, test=test_cost, train=training_cost))
            epoch = float(i) / n_training_samples
            epoch_scores.append((epoch, training_cost, test_cost))
        predictor.train(training_data[[i % n_training_samples]], training_target[[i % n_training_samples]])

    # Plot
    epochs, training_costs, test_costs = zip(*epoch_scores)
    plt.plot(epochs, np.array([training_costs, test_costs]).T)
    plt.xlabel('epoch')
    plt.ylabel('cost')
    plt.legend(['Training Cost', 'Test Cost'])
    plt.title("Learning Curve")
    plt.ion()
    plt.show()

    return {'training_cost': training_cost, 'test_cost': test_cost}


demo_linear_regression.add_variant('fast-learn', eta=0.01)
demo_linear_regression.add_variant('large_input_space', n_in=1000)


if __name__ == "__main__":
    # Open a menu that allows you to run experiments and view old ones.
    demo_linear_regression.browse(display_format="flat")





import logging

from artemis.general.numpy_helpers import get_rng
from artemis.general.should_be_builtins import memoize, bad_value
import numpy as np

from six.moves import xrange

__author__ = 'peter'

# Note - this module used to be called math, but it somehow results in a numpy import error
# due to some kind of name conflict with another module called math.

sigm = lambda x: 1/(1+np.exp(-x))


def cummean(x, axis = None):
    """
    Cumulative mean along axis
    :param x: An array
    :param axis: The axis
    :return: An array of the same shape
    """
    if axis is None:
        assert isinstance(x, list) or x.ndim == 1, 'You must specify axis for a multi-dimensional array'
        axis = 0
    elif axis < 0:
        axis = x.ndim+axis
    x = np.array(x)
    normalizer = np.arange(1, x.shape[axis]+1).astype(float)[(slice(None), )+(None, )*(x.ndim-axis-1)]
    return np.cumsum(x, axis)/normalizer


def cumvar(x, axis = None, sample = True):
    """
    :return: Cumulative variance along axis
    """
    if axis is None:
        assert isinstance(x, list) or x.ndim == 1, 'You must specify axis for a multi-dimensional array'
        axis = 0
    if not isinstance(x, np.ndarray):
        x = np.array(x)
    ex_2 = cummean(x, axis=axis)**2
    e_x2 = cummean(x**2, axis=axis)
    var = e_x2-ex_2
    if sample and x.shape[axis] > 1:
        var *= x.shape[axis]/float(x.shape[axis]-1)
    return var


@memoize
def binary_permutations(n_bits):
    """
    Given some number of bits, return a shape (2**n_bits, n_bits) boolean array containing every permoutation
    of those bits as a row.
    :param n_bits: An integer number of bits
    :return: A shape (2**n_bits, n_bits) boolean array containing every permoutation
        of those bits as a row.
    """
    return np.right_shift(np.arange(2**n_bits)[:, None], np.arange(n_bits-1, -1, -1)[None, :]) & 1


def softmax(x, axis = None):
    """
    The softmax function takes an ndarray, and returns an ndarray of the same size,
    with the softmax function applied along the given axis.  It should always be the
    case that np.allclose(np.sum(softmax(x, axis), axis)==1)
    """
    if axis is None:
        assert x.ndim==1, "You need to specify the axis for softmax if your data is more thn 1-D"
        axis = 0
    x = x - np.max(x, axis=axis, keepdims=True)  # For numerical stability - has no effect mathematically
    expx = np.exp(x)
    return expx/np.sum(expx, axis=axis, keepdims=True)


def expected_sigm_of_norm(mean, std, method = 'probit'):
    """
    Approximate the expected value of the sigmoid of a normal distribution.

    Thanks go to this guy:
    http://math.stackexchange.com/questions/207861/expected-value-of-applying-the-sigmoid-function-to-a-normal-distribution

    :param mean: Mean of the normal distribution
    :param std: Standard Deviation of the normal distribution
    :return: An approximation to Expectation(sigm(N(mu, sigma**2)))
    """
    from scipy.stats import norm
    if method == 'maclauren-2':
        eu = np.exp(-mean)
        approx_exp = 1/(eu+1) + 0.5*(eu-1)*eu/((eu+1)**3) * std**2
        return np.minimum(np.maximum(approx_exp, 0), 1)

    elif method == 'maclauren-3':
        eu = np.exp(-mean)
        approx_exp = 1/(eu+1) + \
            0.5*(eu-1)*eu/((eu+1)**3) * std**2 + \
            (eu**3-11*eu**2+57*eu-1)/((8*(eu+1))**5) * std**4
        return np.minimum(np.maximum(approx_exp, 0), 1)
    elif method == 'probit':
        return norm.cdf(mean/np.sqrt(2.892 + std**2))
    else:
        raise Exception('Method "%s" not known' % method)


l1_error = lambda x1, x2: np.mean(np.abs(x1-x2), axis = -1)


def normalize(x, axis=None, degree = 2, avoid_nans = False):
    """
    Normalize array x.
    :param x: An array
    :param axis: Which axis to normalize along
    :param degree: Degree of normalization (1 for L1-norm, 2 for L2-norm, etc)
    :param avoid_nans: If, along an axis, there is a norm of zero, then normalize this to a uniform vector (instead of nans).
    :return: An array the same shape as x, normalized along the given axis
    """
    assert degree in (1, 2), "Give me a reason and I'll give you more degrees"

    if degree == 1:
        z = np.sum(np.abs(x.astype(np.float)), axis = axis, keepdims=True)
    else:
        z = np.sum(x**degree, axis = axis, keepdims=True)**(1./degree)
    normed = x/z
    if avoid_nans:
        uniform_vector_value = (1./x.shape[axis])**(1./degree)
        normed[np.isnan(normed)] = uniform_vector_value
    return normed


def mode(x, axis = None, keepdims = False):
    from scipy.stats import mode as sp_mode
    mode_x, _ = sp_mode(x, axis = axis)
    if not keepdims:
        mode_x = np.take(mode_x, 0, axis = axis)
    return mode_x


def cummode(x, weights = None, axis = 1):
    """
    Cumulative mode along an axis.  Ties give priority to the first value to achieve the
    given count.
    """
    import weave  # ONLY WORKS IN PYTHON 2.X !!!

    assert x.ndim == 2 and axis == 1, 'Only implemented for a special case!'
    all_values, element_ids = np.unique(x, return_inverse=True)
    n_unique = len(all_values)
    element_ids = element_ids.reshape(x.shape)
    result = np.zeros(x.shape, dtype = int)
    weighted = weights is not None
    if weighted:
        assert x.shape == weights.shape
    counts = np.zeros(n_unique, dtype = float if weighted else int)
    code = """
    bool weighted = %s;
    int n_samples = Nelement_ids[0];
    int n_events = Nelement_ids[1];
    for (int i=0; i<n_samples; i++){
        float maxcount = 0;
        int maxel = -1;

        for (int k=0; k<n_unique; k++)
            counts[k] = 0;

        for (int j=0; j<n_events; j++){
            int ix = i*n_events+j;
            int k = element_ids[ix];
            counts[k] += weighted ? weights[ix] : 1;
            if (counts[k] > maxcount){
                maxcount = counts[k];
                maxel = k;
            }
            result[ix]=maxel;
        }
    }
    """ % ('true' if weighted else 'false')
    weave.inline(code, ['element_ids', 'result', 'n_unique', 'counts', 'weights'], compiler = 'gcc')
    mode_values = all_values[result]
    return mode_values


def recent_moving_average(x, axis = 0):
    """
    Fast computation of recent moving average, where

        frac = 1/sqrt(t)
        a[t] = (1-frac)*a[t-1] + frac*x[t]
    """

    import weave  # ONLY WORKS IN PYTHON 2.X !!!
    if x.ndim!=2:
        y = recent_moving_average(x.reshape(x.shape[0], x.size//x.shape[0]), axis=0)
        return y.reshape(x.shape)

    assert x.ndim == 2 and axis == 0, 'Only implemented for a special case!'
    result = np.zeros(x.shape)
    code = """
    int n_samples = Nx[0];
    int n_dim = Nx[1];
    for (int i=0; i<n_dim; i++)
        result[i] = x[i];
    int ix=n_dim;
    for (int t=1; t<n_samples; t++){
        float frac = 1./sqrt(t+1);
        for (int i=0; i<n_dim; i++){
            result[ix] = (1-frac)*result[ix-n_dim] + frac*x[ix];
        }
        ix += 1;
    }
    """
    weave.inline(code, ['x', 'result'], compiler = 'gcc')
    return result


def angle_between(a, b, axis=None, in_degrees = False):
    """
    Return the angle between two vectors a and b, in radians.  Raise an exception if one is a zero vector
    :param a: A vector
    :param b: A vector the same size as a
    :return: The angle between these vectors, in radians.

    Credit to Pace: http://stackoverflow.com/questions/2827393/angles-between-two-n-dimensional-vectors-in-python
    """
    cos_dist = cosine_distance(a, b, axis=axis)
    angle = np.arccos(cos_dist)
    if in_degrees:
        angle = angle * 180/np.pi
    return angle


def cosine_distance(a, b, axis=None):
    """
    Return the cosine distance between two vectors a and b.  Raise an exception if one is a zero vector
    :param a: An array
    :param b: Another array of the same shape
    :return: The cosine distance between a and b, reduced along the given axis.

    Credit to Pace: http://stackoverflow.com/questions/2827393/angles-between-two-n-dimensional-vectors-in-python
    """
    a = np.array(a) if not isinstance(a, np.ndarray) else a
    b = np.array(b) if not isinstance(b, np.ndarray) else b
    if not a.dtype==float:
        a=a.astype(float)
    if not b.dtype==float:
        b=b.astype(float)
    if axis is None:
        a = a.ravel()
        b = b.ravel()
        axis = 0
    assert a.shape[-1]==b.shape[-1]
    cosine_distance = (a*b).sum(axis=axis)/np.sqrt((a**2).sum(axis=axis) * (b**2).sum(axis=axis))
    # For numerical resons, we might get values outside [-1, 1] here, so we truncate:
    cosine_distance = np.minimum(cosine_distance, 1)
    cosine_distance = np.maximum(cosine_distance, -1)
    return cosine_distance


def degrees_between(a, b):
    return angle_between(a, b, in_degrees=True)


def magnitude_ratio(a, b):
    """
    Return the ratio of the L2-magnitudes of each vector
    :param a: A vector
    :param b: Another vector of the same size
    :return: The ratio |a_mag
    """
    assert a.ndim == 1 and a.shape==b.shape
    a_mag = np.sqrt(np.sum(a**2))
    b_mag = np.sqrt(np.sum(b**2))
    d_magnitude = a_mag/b_mag
    return d_magnitude


def is_parallel(a, b, angular_tolerance = 1e-7):
    """
    Test whether two vectors are parallel to within a given tolerance.
    Throws an exception for zero-vectors.

    :param a: A vector
    :param b: A vector the same size as a
    :param angular_tolerance: The tolerance, in radians.
    :return: A boolean, indicating that the vectors are parallel to within the specified tolerance.
    """
    assert 0 <= angular_tolerance <= 2*np.pi, "It doesn't make sense to specity an angular tolerance outside of [0, 2*pi].  Why are you doing this?"
    angle = angle_between(a, b)
    return angle < angular_tolerance


def align_curves(xs, ys, n_bins='median', xrange = ('min', 'max'), spacing = 'lin'):
    """
    Given multiple curves with different x-coordinates, interpolate so that each has the same x points.

    :param xs: A length-N list of sorted vectors containing the x-coordinates of each curve
    :param ys: A length-N list of vectors containing the corresponding y-coordinates
    :param n_bins: Number of points to make along new x-axis.  'median' to use the median number of points in the curves.
    :param xrange: 2-tuple indicating range of x-axis to span.  'min' indicates "minimum across curves", As with 'max'.
    :param spacing: Either 'lin' or 'log', depenting on whether you want your interpolation points spaced linearly or
        logarithmically.
    :return: (new_xs, new_ys).
        new_xs is a (n_bins, ) curve indicating the new x-locations.
        new_ys is a (N, n_bins)
    """
    assert spacing in ('lin', 'log')
    assert len(xs)==len(ys)
    assert all(len(x)==len(y) for x, y in zip(xs, ys))

    start, stop = xrange
    if start == 'min':
        start = np.min([x[0] for x in xs if len(x)>0])
    if stop == 'max':
        stop = np.max([x[-1] for x in xs if len(x)>0])
    if n_bins == 'median':
        n_bins = int(np.round(np.median([len(x) for x in xs])))

    new_x = np.linspace(start, stop, n_bins) if spacing=='lin' else np.logspace(np.log10(start), np.log10(stop), n_bins)

    new_ys = np.zeros((len(xs), n_bins)) + np.nan

    for x, y, ny in zip(xs, ys, new_ys):
        if len(x)>=2:
            ny[:] = np.interp(x=new_x, xp=x, fp=y, left=np.nan, right=np.nan)

    return new_x, new_ys


def sqrtspace(a, b, n_points):
    """
    :return: Distribute n_points quadratically from point a to point b, inclusive
    """
    return np.linspace(0, 1, n_points)**2*(b-a)+a


def fixed_diff(x, axis=-1, initial_value = 0.):
    """
    Modification of numpy.diff where the first element is compared to the initial value.
    The resulting array has the same shape as x.

    Note that this inverts np.cumsum so that np.cumsum(fixed_diff(x)) == x    (except for numerical errors)

    :param x: An array
    :param axis: Axis along which to diff
    :param initial_value: The initial value agains which to diff the first element along the axis.
    :return: An array of the same shape, representing the difference in x along the axis.
    """

    x = np.array(x, copy=False)

    if axis<0:
        axis = x.ndim+axis

    result = np.empty_like(x)
    initial_indices = (slice(None), )*axis
    result[initial_indices+(slice(1, None), )] = np.diff(x, axis=axis)
    if initial_value == 'first':
        result[initial_indices+(0, )] = 0
    else:
        result[initial_indices+(0, )] = x[initial_indices+(0, )]-initial_value
    return result


def decaying_cumsum(x, memory, axis=-1):

    if axis<0:
        axis = x.ndim+axis
    assert 0 <= memory < 1
    result = np.empty_like(x)
    leading_indices = (slice(None), )*axis
    one_minus_mem = 1-memory
    result[leading_indices+(0, )] = one_minus_mem*x[leading_indices+(0, )]
    for i in xrange(1, x.shape[axis]):
        result[leading_indices+(i, )] = memory*result[leading_indices+(i-1, )] + one_minus_mem*x[leading_indices+(i, )]
    if np.max(np.abs(result)>1e9):
        print('sdfdsf: {}'.format(np.max(np.abs(x))))

    return result


def point_space(start, stop, n_points, spacing):
    if spacing=='lin':
        values = np.linspace(start, stop, n_points)
    elif spacing=='sqrt':
        values = sqrtspace(start, stop, n_points)
    elif spacing=='log':
        values = np.logspace(np.log10(start), np.log10(stop), n_points)
    else:
        raise NotImplementedError(spacing)
    return values


def geosum(rate, t_end, t_start=0):
    """
    Geometric sum of a series from t_start to t_end

    e.g. geosum(0.5, t_end=4, t_start=2) = 0.5**2 + 0.5**3 + 0.5**4 = 0.375
    """
    return np.where(rate==1, np.array(t_end-t_start+1, copy=False).astype(float), np.array(rate**(t_end+1)-rate**t_start)/(rate-1))


def selective_sum(x, ixs):
    """
    :param x: An nd array
    :param ixs: A tuple of length x.ndim indexing each of the dimensions.
    :return: A scalar sum of all array elements selected by any of the dimensions.

    This is best explained by example:

        a = np.array([[ 0,  1,  2,  3],
        ...           [ 4,  5,  6,  7],
        ...           [ 8,  9, 10, 11],
        ...           [12, 13, 14, 15]])

    If we want to add all elements rows 1, and 2, as well as the column, then we go:
        s = selective_sum(a, [(1,3), 2])

    And we can verify that:
        s == 4+5+6+7 + 12+13+14+15 + 2+10 == 88

    If you don't want to select coluns
    """
    assert x.ndim==len(ixs), 'The dimension of x must match the length of ixs'
    al = (slice(None), )
    selection_mask = np.zeros(x.shape, dtype='bool')
    for i, ix in enumerate(ixs):
        selection_mask[al*i+(ix, )+al*(x.ndim-i-1)] = True
    return (x*selection_mask).sum()

    # Note, we'd like to do this more efficiently, but it gets a little complicated.
    # (we have to add the individual indexes, but subtract the double-counted regions, and then subtract the triple-counted
    # regions, and so on....)
    # return sum(x[al*i+(ix, )+al*(x.ndim-i-1)].sum() for i, ix in enumerate(ixs)) - x[ixs].sum()


def conv_fanout(input_len, kernel_len, conv_mode):
    """
    Note: this is horrific and must be simplified.
    :param input_len:
    :param kernel_len:
    :param conv_mode:
    :return:
    """

    if conv_mode=='full':
        return kernel_len*np.ones(input_len)
    else:
        if conv_mode=='half':
            conv_mode='same'
        left_pad = kernel_len // 2 if conv_mode == 'same' else 0 if conv_mode == 'valid' else conv_mode if isinstance(conv_mode, int) else bad_value(conv_mode)
        right_pad = (kernel_len-1) // 2 if conv_mode == 'same' else 0 if conv_mode == 'valid' else conv_mode if isinstance(conv_mode, int) else bad_value(conv_mode)
        full_range = np.arange(left_pad + input_len + right_pad)
        max_fanout = np.minimum(kernel_len, np.maximum(input_len-kernel_len+1+2*left_pad, 1))
        fanout_over_full_range = np.minimum(max_fanout, np.minimum(full_range+1, full_range[::-1]+1))
        fanout = fanout_over_full_range[left_pad:len(full_range)-right_pad]
        return fanout


def conv2_fanout_map(input_shape, kernel_shape, conv_mode):
    size_y, size_x = input_shape
    k_size_y, k_size_x = kernel_shape
    y_fanout = conv_fanout(input_len = size_y, kernel_len=k_size_y, conv_mode=conv_mode)
    x_fanout = conv_fanout(input_len = size_x, kernel_len=k_size_x, conv_mode=conv_mode)
    fanout_map = y_fanout[:, None] * x_fanout
    return fanout_map


def levenshtein_distance(s1, s2):
    """
    The Levenshtein Distance (a type of edit distance) between strings

    Thank you to Salvador Dali here: https://stackoverflow.com/a/32558749/851699
    :param s1: A string
    :param s2: Another String
    :return: An integer distance.
    """
    if len(s1) > len(s2):
        s1, s2 = s2, s1
    distances = range(len(s1) + 1)
    for i2, c2 in enumerate(s2):
        distances_ = [i2+1]
        for i1, c1 in enumerate(s1):
            if c1 == c2:
                distances_.append(distances[i1])
            else:
                distances_.append(1 + min((distances[i1], distances[i1 + 1], distances_[-1])))
        distances = distances_
    return distances[-1]


def onehotvector(ix, length):
    """
    Create a one-hot vector of length length with element ix 1 (and the rest 0)
    :param ix: The element to be "hot".  Ot a vector of elememts.
    :param length: The total length of the vector.
    :return: If ix is scalar, a single vector.
        If ix is a vector, a (len(ix), length) array where each row is a one-hot vector for an element of ix.
    """
    if isinstance(ix, int):
        v = np.zeros(length)
        v[ix] = 1
    else:
        v = np.zeros((len(ix), length))
        v[np.arange(len(ix)), ix] = 1
    return v


def proportional_random_assignment(length, split, rng):
    """
    Generate an integer array of the given length, with elements randomly assigned to 0...len(split), with
    frequency of elements with value i proporational to split[i].

    This is useful for splitting training/test sets.  e.g.

        n_samples = 1000
        x = np.random.randn(n_samples, 4)
        y = np.random.randn(n_samples)
        subsets = proportional_random_assignment(n_samples, split=0.7, rng=1234)
        x_train = x[subsets==0]
        y_train = y[subsets==0]
        x_test = x[subsets==1]
        y_test = y[subsets==1]

    :param length: The length of the output array
    :param split: Either a list of ratios to assign to each group (must add to <1), or a single float in (0, 1),
        which will indicate that we split into 2 groups.
    :param rng: A random number generator or seed.
    :return: An integer array.
    """
    rng = get_rng(rng)
    if isinstance(split, float):
        split = [split]
    assert 0<=np.sum(split)<=1, "The sum of elements in split: {} must be in [0, 1].  Got {}".format(split, np.sum(split))
    arr = np.zeros(length, dtype=int)
    cut_points = np.concatenate([np.round(np.cumsum(split)*length).astype(int), [length]])
    scrambled_indices = rng.permutation(length)
    for i, (c_start, c_end) in enumerate(zip(cut_points[:-1], cut_points[1:])):
        arr[scrambled_indices[c_start:c_end]] = i+1  # Note we skip zero since arrays already inited to 0
    return arr


def argmaxnd(x):
    ix = np.argmax(x.flatten())
    return np.unravel_index(ix, dims=x.shape)


def clip_to_sum(vec, total):
    new_vec = np.array(vec)  # Yes this is horribly inefficient but I do not care.
    current_total = np.sum(vec)
    while current_total > total:
        i = np.argmax(new_vec)
        new_vec[i] -= 1
        current_total -= 1
    return new_vec



import textwrap
from collections import OrderedDict

from artemis.general.display import IndentPrint, CaptureStdOut, side_by_side, DocumentWrapper, deepstr, \
    str_with_arrayopts, surround_with_header, sensible_str, format_duration
import numpy as np

_desired = """
aaa
bbb
    ccc
    ddd
        eee
        fff
    ggg
    hhh
iii
jjj
"""


def test_indent_print():

    with CaptureStdOut() as cap:
        print('aaa')
        print('bbb')
        with IndentPrint():
            print('ccc')
            print('ddd')
            with IndentPrint():
                print('eee')
                print('fff')
            print('ggg')
            print('hhh')
        print('iii')
        print('jjj')

    assert '\n'+cap.read() == _desired


str1 = '' + \
    'For those who are led by the Spirit of God are the\n' \
    'children of God.\n' \
    'Romans 8:14'

str2 = '' \
       'But to you who are listening I say: Love your enemies,\n' \
       'do good to those who hate you, bless those who curse you, pray\n' \
       'for those who mistreat you.\n' \
       'Luke 6:27-28'

desired = '' \
    'For those who are led by the Spirit of God are the    But to you who are listening I say: Love your enemies,        \n' \
    'children of God.                                      do good to those who hate you, bless those who curse you, pray\n' \
    'Romans 8:14                                           for those who mistreat you.                                   \n' \
    '                                                      Luke 6:27-28                                                  '


def test_side_by_side():

    print('String 1:\n{}'.format(str1))
    print('String 2:\n{}'.format(str2))

    out = side_by_side([str1, str2])
    print('Side by side:\n{}'.format(out))
    assert out==desired  # Would work but pycharm automatically trims trailing spaces from the strings defined av


def test_document_wrapper():
    """
    Document Wrapper is deals with wrapping text with new lines already present.
    :return:
    """

    str3="0123456789\n0123456789\n01234567890123456789"

    desired3="0123456789\n0123456789\n012345678901\n23456789"

    w1 = textwrap.TextWrapper(width=12, replace_whitespace=False)
    r1 = w1.fill(str3)
    assert r1 != desired3
    assert r1 == "0123456789\n0123456789\n0\n123456789012\n3456789"

    w2 = DocumentWrapper(width=12, replace_whitespace=False)
    r2 = w2.fill(str3)
    assert r2 == desired3


def test_deepstr():

    obj = {'a': np.arange(100).reshape(10, 10), 'bbbb': [1, 3, np.arange(6).reshape(2, 3), ('xx', 'yy')]}
    obj['c'] = obj['bbbb']
    string_desc = deepstr(obj)
    print(string_desc)

    obj = {}
    string_desc = deepstr(obj)
    print(string_desc)
    # For now, no assertions, because string contains IDS which will always change.  We can come up with some way to do this later with regular experessions if needed.


def test_str_with_arrayopts():

    a = np.arange(1, 6, 1./3)
    assert str_with_arrayopts(a, float_format='.3g', threshold=None) == '[1 1.33 1.67 2 2.33 2.67 3 3.33 3.67 4 4.33 4.67 5 5.33 5.67]'
    assert str_with_arrayopts(a, float_format='.3g', threshold=5) == '[1 1.33 1.67 ..., 5 5.33 5.67]'  # Yeah there's a minimum of 3 on each end.

    b = a.reshape(3, 5)
    assert str_with_arrayopts(b, float_format='.3g', threshold=5) == '[[1 1.33 1.67 2 2.33]\n [2.67 3 3.33 3.67 4]\n [4.33 4.67 5 5.33 5.67]]'


def test_surround_with_header():

    a = surround_with_header('abcd', width=40)
    assert len(a)==40
    b = surround_with_header('abcde', width=40)
    assert len(b)==40

    a = surround_with_header('abcd', width=41)
    assert len(a)==41
    b = surround_with_header('abcde', width=41)
    assert len(b)==41

    a = surround_with_header('abcd', width=2)
    assert len(a)==6


def test_nested_capture():

    with CaptureStdOut() as cap1:
        print('a')
        with CaptureStdOut() as cap2:
            print('b')
        print('c')

    assert cap2.read()=='b\n'
    assert cap1.read()=='a\nb\nc\n'


def test_capture_prefix():

    with CaptureStdOut() as cap1:
        print('a')
        with CaptureStdOut(prefix='abc:') as cap2:
            print('b')
        print('c')

    print ('Done')
    assert cap2.read()=='b\n'
    assert cap1.read()=='a\nabc:b\nc\n'


def test_sensible_str():

    a = [1, 2, 3]
    stra = sensible_str(a)
    assert stra=='[1,2,3]'

    a = OrderedDict([('a', [1,2,3]), ('b', (3, list(range(20)))), ('c', np.arange(20).reshape(4, 5)), ('d', np.arange(4).reshape(2, 2))])
    stra = sensible_str(a, size_limit=4, compact=True)
    assert stra=='OrderedDict([a:[1,2,3],b:(3,<len20-list>),c:<(4,5)ndarray>,d:ndarray([[01],[23]])])'


def test_format_duration():
    assert format_duration(0.0025) == '2.5ms'
    assert format_duration(0.25) == '250.0ms'
    assert format_duration(25) == '25.0s'
    assert format_duration(250) == '0:04:10'
    assert format_duration(2500) == '0:41:40'
    assert format_duration(25000) == '6:56:40'
    assert format_duration(250000) == '2d,21:26:40'
    assert format_duration(2500000) == '28d,22:26:40'
    assert format_duration(25000000) == '289d,8:26:40'
    assert format_duration(250000000) == '2893d,12:26:40'


if __name__ == '__main__':
    test_indent_print()
    test_side_by_side()
    test_document_wrapper()
    test_deepstr()
    test_str_with_arrayopts()
    test_surround_with_header()
    test_nested_capture()
    test_capture_prefix()
    test_sensible_str()
    test_format_duration()



from collections import OrderedDict
import itertools
import numpy as np
from abc import ABCMeta, abstractmethod
from artemis.general.display import arraystr, indent_string
from artemis.general.should_be_builtins import izip_equal
import sys


class CollectionIsEmptyException(Exception):
    pass


class InvalidKeyError(KeyError):
    pass


class InvalidInitializerError(Exception):
    pass


class InitializerTooShallowError(Exception):
    pass


class UniversalCollection(object):

    __metaclass__ = ABCMeta

    @abstractmethod
    def __contains__(self, item):
        raise NotImplementedError()

    @abstractmethod
    def keys(self):
        raise NotImplementedError()

    @abstractmethod
    def values(self):
        raise NotImplementedError()

    @abstractmethod
    def has_key(self, key):
        raise NotImplementedError()

    @abstractmethod
    def __getitem__(self, ix):
        raise NotImplementedError()

    @abstractmethod
    def __setitem__(self, key, value):
        raise NotImplementedError()

    def __eq__(self, other):
        return self.to_struct() == (other.to_struct() if isinstance(other, UniversalCollection) else other)

    @abstractmethod
    def to_struct(self):
        raise NotImplementedError()

    @abstractmethod
    def __len__(self):
        raise NotImplementedError()

    @abstractmethod
    def __iter__(self):
        raise NotImplementedError()

    def items(self):
        return zip(self.keys(), self.values())

    @staticmethod
    def from_first_key(key):
        if key is next or isinstance(key, int):
            obj = DynamicSequence()
        else:
            obj = UniversalOrderedStruct()
        return obj

    @classmethod
    def from_struct(cls, struct):
        if isinstance(struct, (list, tuple)):
            return DynamicSequence(struct)
        elif isinstance(struct, dict):
            return UniversalOrderedStruct(struct)
        elif struct is None or isinstance(struct, EmptyCollection):
            return EmptyCollection()
        else:
            raise InvalidInitializerError("Don't know how to load a Universal Collection from initializer: {}".format(struct))

    def map(self, f):
        new_obj = self.__class__()
        for k, v in self.items():
            new_obj[k] = f(v)
        return new_obj


class EmptyCollection(UniversalCollection):

    def has_key(self, key):
        return False

    def __getitem__(self, ix):
        if isinstance(ix, slice):
            return []
        else:
            raise CollectionIsEmptyException()

    def __setitem__(self, ix, val):
        raise CollectionIsEmptyException()

    def __contains__(self, item):
        return False

    def __len__(self):
        return 0

    def keys(self):
        return []

    def values(self):
        return []

    def to_struct(self):
        return None

    def items(self):
        return []

    @staticmethod
    def from_struct(struct):
        if struct is not None:
            raise CollectionIsEmptyException()

    def __iter__(self):
        return iter([])


class DynamicSequence(list, UniversalCollection):
    """
    This mostly behaves like a list, except
    - It's also extended to have the key, value methods of an OrderedDict.
    - It also supports numpy-like indexing with a list of integers.
    - It allows you to assign with the built-in "next" as an index.  Eg "a[next] = 3" which means
      "append 3 to the list"
    """

    def values(self):
        return self

    def keys(self):
        return range(len(self))

    def __getitem__(self, ix):
        if isinstance(ix, slice):
            return DynamicSequence(list.__getitem__(self, ix))
        elif isinstance(ix, (list, tuple)):
            return DynamicSequence((list.__getitem__(self, i) for i in ix))
        else:
            try:
                return list.__getitem__(self, ix)
            except TypeError:
                raise InvalidKeyError('You tried getting index "{}" from a {}, but {} object can only be indexed by ints, slices, numeric lists'.format(ix, type(self).__name__, type(self).__name__))

    def __setitem__(self, ix, val):
        if ix is next:
            self.append(val)
        elif isinstance(ix, int):
            if ix==len(self):
                self.append(val)
            else:
                try:
                    list.__setitem__(self, ix, val)
                except IndexError:
                    raise InvalidKeyError('If you assign to a DynamicSequence, the index must be no greater than the length of the sequence.  Got index {} for length {}'.format(ix, len(self)))
        else:
            raise InvalidKeyError('Cannot index a DynamicSequence with non-integer index: {}'.format(ix))

    def has_key(self, key):
        try:
            self[key]
            return True
        except IndexError:
            return False

    def items(self):
        return zip(range(len(self)), self)

    def to_struct(self):
        return list(self)

    @classmethod
    def from_struct(cls, struct):
        assert isinstance(struct, list)
        return DynamicSequence(struct)


class UniversalOrderedStruct(UniversalCollection):
    """
    Mostly like an OrderedDict, but it behaves like a list, in that (for x in struct) and (x in struct) looks over
    values, not keys.
    """

    def __init__(self, *initializer):
        if len(initializer)==1 and type(initializer[0]) is dict:
            d = initializer[0]
            initializer = [OrderedDict((k, d[k]) for k in sorted(d.keys()))]

        self._heart = OrderedDict(*initializer)

    def __contains__(self, item):
        return item in self._heart.values()

    def __setitem__(self, key, value):
        if key is next or isinstance(key, int):
            raise InvalidKeyError('This sequence is an {}, and cannot be given {} key: {}'.format(self.__class__.__name__, key.__class__.__name__, key))
        self._heart.__setitem__(key, value)

    def __getitem__(self, selector):
        if isinstance(selector, (list, slice, np.ndarray)):
            if isinstance(selector, slice):
                all_keys = list(self.keys())
                start_index = all_keys.index(selector.start) if selector.start is not None else None
                stop_index = all_keys.index(selector.stop) if selector.stop is not None else None
                keys = all_keys[start_index:stop_index:selector.step]
            else:
                keys = selector
            return UniversalOrderedStruct((k, self[k]) for k in keys)
        else:
            return self._heart.__getitem__(selector)

    def __repr__(self):
        rep = self._heart.__repr__()
        return self.__class__.__name__ + rep[len(OrderedDict.__class__.__name__):]

    def __iter__(self):
        return iter(self._heart.values())

    def __len__(self):
        return self._heart.__len__()

    def has_key(self, key):
        return key in self._heart

    def keys(self):
        return self._heart.keys()

    def values(self):
        return self._heart.values()

    def to_struct(self):
        return self._heart.copy()

    @classmethod
    def from_struct(cls, struct):
        return cls(struct)


class Duck(UniversalCollection):
    """
    A dynamic data structure that makes it easy to store and query entries.  It's a cross between a dict, an array, and
    a list.  It behaves similarly to a numpy array, but does not require that data be regularly sized.

        a = Duck()
        a['a', 'aa1'] = 1
        a['a', 'aa2'] = 2
        a['b', 0, 'subfield1'] = 4
        a['b', 0, 'subfield2'] = 5
        a['b', 1, 'subfield1'] = 6
        a['b', 1, 'subfield2'] = 7

        assert list(a['b', 1, :]) == [6, 7]
        assert a['b', :, 'subfield1'] == [4, 6]

    """
    def __init__(self, initial_struct = None, recurse = False):
        if recurse is True:  # How many layers deep to recurse.  0=just the initial struct, 1=break into initial_struct_keys, ...
            self._struct = UniversalCollection.from_struct(initial_struct).map(
                (lambda x: (Duck(x, recurse=recurse if recurse is True else recurse - 1)
                    if isinstance(x, (dict, list, tuple, UniversalCollection, type(None))) else x)))
        elif isinstance(recurse, int) and recurse>0:  # How many layers deep to recurse.  0=just the initial struct, 1=break into initial_struct_keys, ...
            try:
                self._struct = UniversalCollection.from_struct(initial_struct).map((lambda x: (Duck(x, recurse=recurse - 1))))
            except InitializerTooShallowError:
                raise InitializerTooShallowError('Tried to initialize a Duck by breaking {} levels into an object, but the not all branches of go {} levels deep.  The object: {}'.format(recurse+1, recurse+1, initial_struct))
        elif isinstance(initial_struct, Duck):
            raise Exception('fdsf')
            # self._struct = DictArrayList.from_struct()
        elif isinstance(initial_struct, UniversalCollection):
            self._struct = initial_struct
        else:
            try:
                self._struct = UniversalCollection.from_struct(initial_struct)
            except InvalidInitializerError:
                raise InitializerTooShallowError('Initializer {} has no depth, and so cannot be used to initialize a Duct.'.format(initial_struct))

    def __setitem__(self, indices, value):
        """
        :param indices: A tuple of indices (as you would use to access an element in an array)
        :param value: An arbitrary object to place at those indices.  If it is another Duck, it will be considered a child node of this Duck.
        """
        key_chain = indices if isinstance(indices, tuple) else (indices,)
        if len(key_chain)==0:  # e.g.  a[()]=4   Typically this happens when assigning to a slice.  ie a[:]=[1,2,3] becomes a[()]=Duck([1,2,3]).  Similar syntax in numpy
            self._struct = value._struct if isinstance(value, Duck) else value  # Possibly weird edge cases could arise here...
        elif any(isinstance(k, slice) or k is Ellipsis for k in key_chain):  # When training elements are slices, ellipses.
            open_keys = [s==slice(None) or s is Ellipsis for s in key_chain]
            first_open_key_index = open_keys.index(True, )
            if not all(open_keys[first_open_key_index:]):
                raise InvalidKeyError("Currently, sliced assignment is only supported for the final indices.  Got {}".format(key_chain))
            self.__setitem__(key_chain[:first_open_key_index], Duck(value, recurse=True if key_chain[-1] is Ellipsis else len(key_chain) - first_open_key_index - 1))
        else:  # The common case: an unsliced assignment
            first_key, subkeys = key_chain[0], key_chain[1:]
            try:
                if len(subkeys)==0:  # You assign directly to this object
                    self._struct[first_key] = value
                else:  # This is just getting a container for an assignment
                    assert not isinstance(first_key, slice), 'Sliced assignment currently not supported'
                    if not self._struct.has_key(first_key):
                        self._struct[first_key] = Duck()
                    self._struct[first_key][subkeys] = value
            except CollectionIsEmptyException:
                self._struct = UniversalCollection.from_first_key(first_key)
                self.__setitem__(indices, value)

    def __getitem__(self, indices):
        """
        :param indices: A tuple of indices (as you would use to access an element in an array)
        :return: Either
            a) A New Duck (if the selectors specifiy a range of elements in the current Duck
            b) An element at a leaf node (if the selectors specify a partucular item)
        """
        if indices == ():
            return self
        first_selector, indices = (indices[0], indices) if isinstance(indices, tuple) else (indices, (indices,))

        if self._struct is None:
            raise Exception('No value has ever been assigned to this Duck, but you are trying to extract index {} from it.'.format(indices))

        try:
            new_substruct = self._struct[first_selector]
            if isinstance(new_substruct, UniversalCollection) and not isinstance(new_substruct, Duck):  # This will happen if the selector is a slice or something...
                new_substruct = Duck(new_substruct, recurse=False)
            if len(indices)==1:  # Case 1: Simple... this is the last selector, so we can just return it.
                return new_substruct
            else:  # Case 2:
                # assert isinstance(new_substruct, ArrayStruct), 'You are selecting with {}, which indicates a depth of {}, but the structure has no more depth.'.format(list(':' if s==slice(None) else s for s in selectors), len(selectors))
                if isinstance(first_selector, (list, np.ndarray, slice)):  # Sliced selection, with more sub-indices
                    return new_substruct.map(lambda x: x.__getitem__(indices[1:]))
                else:  # Simple selection, with more sub-indices
                    return new_substruct[indices[1:]]
        except CollectionIsEmptyException:
            raise Exception('No value has ever been assigned to this Duck, but you are trying to extract index {} from it.'.format(indices))

    def deepvalues(self):
        return [(v.deepvalues() if isinstance(v, Duck) else v.values() if isinstance(v, UniversalCollection) else v) for v in self._struct]

    def map(self, f):
        return Duck(self._struct.map(f))

    def _first_element_keys(self):

        these_keys = list(self.keys())
        try:
            first_el = next(iter(self._struct))  # Try to get the keys of the first element
        except StopIteration:
            return (these_keys, )  # If there is no first element, there're no keys to get
        if isinstance(first_el, Duck):
            return (these_keys, )+first_el._first_element_keys()
        else:
            return (these_keys, )

    def to_array_and_keys(self):
        keys = self._first_element_keys()
        first = True
        key_indices = tuple(range(len(k)) for k in keys)
        for numeric_ix, key_ix in izip_equal(itertools.product(*key_indices), itertools.product(*keys)):
            if first:
                this_data = np.array(self[key_ix])
                arr = np.zeros(tuple(len(k) for k in keys)+this_data.shape, dtype=this_data.dtype)
                arr[numeric_ix] = this_data
                first = False
            else:
                arr[numeric_ix] = self[key_ix]
        return keys, arr

    def to_array(self):
        """
        Turn the object into a numpy array.  This will fail unless all sub-branches have the same structure (ie
        the object is non-ragged)
        :return: A numpy array
        """
        _, arr = self.to_array_and_keys()
        return arr

    def arrayify_axis(self, axis, subkeys = (), inplace=False):
        """
        Take a particular axis (depth) of this object, concatenate it into an array for each leaf node.  This requires
        that all values along this axis have the same structure.  Best understood by the following example:

            a = Duck()
            a[0, 'x'] = 1
            a[0, 'y'] = 2
            a[1, 'x'] = 3
            a[1, 'y'] = 4
            b = a.arrayify_axis(axis=0)
            assert np.array_equal(b['x'], [1, 3])
            assert np.array_equal(b['y'], [2, 4])

        :param axis: The depth at which you'd like to concatenate.
        :param subkeys: A sub-branch of struct to do this along (ie. only do it on branch ('subkey', 'subsubkey'))
        :param inplace: Do the operation in-place (ie change this object (saves you from copying a new object))
        :return: A new Duck, with the given axis shifted to the end.
        """
        b = Duck.from_struct(self.to_struct()) if not inplace else self
        if not isinstance(subkeys, tuple):
            subkeys = (subkeys,)
        indexing_keys = list(self[subkeys]._first_element_keys())
        indexing_keys[axis - len(subkeys)] = (slice(None),)
        indexing_keys = tuple(indexing_keys)
        assigning_keys = list(indexing_keys)
        del assigning_keys[axis - len(subkeys)]
        if len(assigning_keys)==0:
            assert len(indexing_keys)==1
            substruct = np.array(self[subkeys+indexing_keys[0]])
        else:
            substruct = Duck()
            for src_key, dest_key in izip_equal(itertools.product(*indexing_keys), itertools.product(*assigning_keys)):
                substruct[dest_key] = np.array(self[subkeys + src_key])
        b[subkeys] = substruct
        return b

    def to_struct(self):
        return self._struct.map(lambda x: x.to_struct() if isinstance(x, UniversalCollection) else x).to_struct()

    @classmethod
    def from_struct(cls, struct):
        return cls(initial_struct=struct, recurse=True)

    def __contains__(self, item):
        return self._struct.__contains__(item)

    def __len__(self):
        return self._struct.__len__()

    def __eq__(self, other):
        """
        We define equality as both keys and values matching for the full depth.
        :param other: Another Duck, or nested object.
        :return: True if equal, False if not.
        """
        if len(self) != len(other):
            return False
        if not len(self)==len(other):
            return False
        if not isinstance(other, (UniversalCollection, list, OrderedDict, np.ndarray)):
            return False
        for (our_key, our_value) in self.items():
            if self[our_key] != other[our_key]:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __iter__(self):
        return self._struct.__iter__()

    def open(self, *ixs):
        self[ixs] = Duck()
        ixs = tuple(-1 if ix is next else ix for ix in ixs)
        return self[ixs]

    def has_key(self, *key_chain):
        return self._struct.has_key()

    def keys(self, depth=None):
        if depth is None:
            for k in self._struct.keys():
                yield k
        elif depth=='full' or depth>0:
            if isinstance(depth, str):
                assert depth=='full'
                depth = sys.maxsize
            for k, v in self._struct.items():
                if isinstance(v, Duck) and depth>1:
                    for subkey in v.keys(depth=depth-1):
                        yield (k, )+subkey
                else:
                    yield (k, )
        else:
            raise Exception('"depth" can be "full", None, or an integer >= 1.  Got {}'.format(depth))

    def values(self, depth=None):
        for k in self.keys(depth=depth):
            yield self[k]

    def items(self, depth=None):
        for k in self.keys(depth=depth):
            yield k, self[k]

    def __str__(self, max_key_len=4):
        keys = list(self.keys())
        if len(keys)>max_key_len:
            inner_description = [str(k) for k in keys[:max_key_len-1]]+['...']
        else:
            inner_description = [str(k) for k in keys]
        return '<{} with {} keys: {}>'.format(self.__class__.__name__, len(keys), inner_description)

    def description(self, max_expansion=4, _skip_intro=False):
        """
        :param max_expansion: Maximum amount to break into structures
        :param _skip_intro: For internal use in recursion
        :return: A string describing the structure of this Duck.
        """
        full_string = '' if _skip_intro else (str(self) + '')
        key_value_string = ''
        for i, (k, v) in enumerate(self._struct.items()):
            if i>max_expansion:
                key_value_string += '\n(... Omitting {} of {} elements ...)'.format(len(self._struct)-max_expansion, len(self._struct))
            else:
                if isinstance(v, Duck):
                    value_string = v.description(max_expansion=max_expansion, _skip_intro=True)
                elif isinstance(v, np.ndarray):
                    value_string = arraystr(v)
                else:
                    value_string = str(v)
                key_value_string += '\n{}: {}'.format(str(k), value_string)
            full_string += key_value_string

            if i>max_expansion:
                break
        return ('' if _skip_intro else (str(self) + '')) + indent_string(key_value_string, indent='| ', include_first=False)



from six.moves import input

from artemis.fileman.directory_crawl import DirectoryCrawler
from artemis.general.display import surround_with_header


def select_item_ui(item_name_list, prompt = 'Select Items>>'):

    print('\n'.join('{}: {}'.format(i, item) for i, item in enumerate(item_name_list)))
    cmd = input(prompt)
    if cmd=='all':
        ixs = range(item_name_list)
    elif '-' in cmd:
        start, end = cmd.split('-')
        ixs = range(int(start), int(end)+1)
    else:
        ixs = [int(cmd)]
    return ixs




import numpy as np
import pytest
from six.moves import xrange

from artemis.ml.tools.processors import RunningAverage, RecentRunningAverage

__author__ = 'peter'


def test_running_average():

    inp = np.arange(5)
    processor = RunningAverage()
    out = [processor(el) for el in inp]
    assert out == [0, 0.5, 1, 1.5, 2]
    assert np.array_equal(out, RunningAverage.batch(inp))

    inp = np.random.randn(10, 5)
    processor = RunningAverage()
    out = [processor(el) for el in inp]
    assert all(np.allclose(out[i], np.mean(inp[:i+1], axis = 0)) for i in xrange(len(inp)))


@pytest.mark.skipif(True, reason='Depends on weave, which is deprecated for python 3')
def test_recent_running_average():

    inp = np.arange(5)
    processor = RecentRunningAverage()
    out = [processor(el) for el in inp]
    out2 = processor.batch(inp)
    assert np.allclose(out, out2)
    assert np.allclose(out, [0.0, 0.7071067811865475, 1.4535590291019362, 2.226779514550968, 3.019787823462811])

    inp = np.random.randn(10, 5)
    processor = RunningAverage()
    out = [processor(el) for el in inp]
    out2 = processor.batch(inp)
    assert np.allclose(out, out2)


if __name__ == '__main__':

    # test_running_average()
    test_recent_running_average()


import pytest

from artemis.experiments.decorators import ExperimentFunction, experiment_function
from artemis.experiments.experiment_record_view import get_oneline_result_string, print_experiment_record_argtable, \
    compare_experiment_records, get_record_invalid_arg_string
from artemis.experiments.experiments import experiment_testing_context, clear_all_experiments
from artemis.general.display import CaptureStdOut, assert_things_are_printed


def display_it(result):
    print(str(result) + 'aaa')


def one_liner(result):
    return str(result) + 'bbb'


def compare_them(results):
    print(', '.join('{}: {}'.format(k, results[k]) for k in sorted(results.keys())))


@ExperimentFunction(display_function=display_it, one_liner_function=one_liner, comparison_function=compare_them)
def my_xxxyyy_test_experiment(a=1, b=2):

    if b==17:
        raise Exception('b should never be 17')

    print('xxx' if a==1 else 'yyy')
    return a+b


my_xxxyyy_test_experiment.add_variant('a2', a=2)
my_xxxyyy_test_experiment.add_variant(b=17)


def test_experiments_function_additions():

    with experiment_testing_context():

        for rec in my_xxxyyy_test_experiment.get_variant_records(flat=True):
            rec.delete()

        r1=my_xxxyyy_test_experiment.run()
        r2=my_xxxyyy_test_experiment.get_variant('a2').run()
        with pytest.raises(Exception):
            my_xxxyyy_test_experiment.get_variant(b=17).run()
        r3 = my_xxxyyy_test_experiment.get_variant(b=17).get_latest_record()

        assert r1.get_log() == 'xxx\n'
        assert r2.get_log() == 'yyy\n'

        assert get_oneline_result_string(my_xxxyyy_test_experiment.get_latest_record()) == '3bbb'
        assert get_oneline_result_string(my_xxxyyy_test_experiment.get_variant('a2').get_latest_record()) == '4bbb'
        assert get_oneline_result_string(my_xxxyyy_test_experiment.get_variant(b=17).get_latest_record()) == '<No result has been saved>'

        with CaptureStdOut() as cap:
            my_xxxyyy_test_experiment.show(my_xxxyyy_test_experiment.get_latest_record())
        assert cap.read() == '3aaa\n'

        with CaptureStdOut() as cap:
            my_xxxyyy_test_experiment.compare([r1, r2])
        assert cap.read() == 'my_xxxyyy_test_experiment: 3, my_xxxyyy_test_experiment.a2: 4\n'

        print('='*100+'\n ARGTABLE \n'+'='*100)
        print_experiment_record_argtable([r1, r2, r3])

        print('='*100+'\n SHOW \n'+'='*100)
        compare_experiment_records([r1, r2, r3])


def test_experiment_function_ui():

    with experiment_testing_context():
        for existing_record in my_xxxyyy_test_experiment.get_variant_records(flat=True):
            existing_record.delete()

        assert len(my_xxxyyy_test_experiment.get_variant_records(flat=True)) == 0

        my_xxxyyy_test_experiment.browse(raise_display_errors=True, command='run all', close_after=True)
        assert len(my_xxxyyy_test_experiment.get_variant_records()) == 3

        import time
        time.sleep(0.1)

        with assert_things_are_printed(min_len=1200, things=['Common Args', 'Different Args', 'Result', 'a=1, b=2', 'a=2, b=2', 'a=1, b=17']):
            my_xxxyyy_test_experiment.browse(raise_display_errors=True, command='argtable all', close_after=True)

        with assert_things_are_printed(min_len=600, things=['my_xxxyyy_test_experiment: 3', 'my_xxxyyy_test_experiment.a2: 4']):
            my_xxxyyy_test_experiment.browse(raise_display_errors=True, command='compare all -r', close_after=True)

        with assert_things_are_printed(min_len=1300, things=['Result', 'Logs', 'Ran Succesfully', 'Traceback']):
            my_xxxyyy_test_experiment.browse(raise_display_errors=True, command='show all -o', close_after=True)

        with assert_things_are_printed(min_len=1250, things=['3aaa', '4aaa']):
            my_xxxyyy_test_experiment.browse(raise_display_errors=True, command='show all -r', close_after=True)

        with assert_things_are_printed(min_len=1700, things=['Result', 'Logs', 'Ran Succesfully']):
            my_xxxyyy_test_experiment.browse(raise_display_errors=True, command='show 0 -o', close_after=True)


def test_invalid_arg_text():

    with experiment_testing_context(new_experiment_lib=True):

        @experiment_function
        def my_invalid_arg_test(a=1, b={'c': 3, 'd': 4}):
            return a+b['c']+b['d']

        record = my_invalid_arg_test.run()
        assert get_record_invalid_arg_string(record, recursive=True) == '<No Change>'
        clear_all_experiments()

        @experiment_function
        def my_invalid_arg_test(a=2, b={'c': 3, 'd': 4}):
            return a+b['c']+b['d']

        assert get_record_invalid_arg_string(record, recursive=True) == 'Change: {a:1}->{a:2}'
        clear_all_experiments()

        @experiment_function
        def my_invalid_arg_test(a=2, b={'c': 3, 'd': 2}):
            return a+b['c']+b['d']

        assert get_record_invalid_arg_string(record, recursive=True) == "Change: {a:1,b['d']:4}->{a:2,b['d']:2}"


class MyArgumentObject(object):

    def __init__(self, a=1):
        self.a=a


def test_invalid_arg_text_when_object_arg():

    with experiment_testing_context(new_experiment_lib=True):

        @experiment_function
        def my_unhashable_arg_test(a=MyArgumentObject(a=3)):
            return a.a+2

        record = my_unhashable_arg_test.run()
        assert record.get_result() == 5

        assert get_record_invalid_arg_string(record, recursive=True) == '<No Change>'

        # ---------------------
        clear_all_experiments()

        @experiment_function
        def my_unhashable_arg_test(a=MyArgumentObject(a=3)):
            return a.a+2

        assert get_record_invalid_arg_string(record, recursive=True) == '<No Change>'

        # ---------------------
        clear_all_experiments()

        @experiment_function
        def my_unhashable_arg_test(a=MyArgumentObject(a=4)):
            return a.a+2

        assert get_record_invalid_arg_string(record, recursive=True) == 'Change: {a.a:3}->{a.a:4}'


def test_simple_experiment_show():

    with experiment_testing_context(new_experiment_lib=True):

        @experiment_function
        def my_simdfsfdsgfs(a=1):

            print('xxxxx')
            print('yyyyy')
            return a+2

        rec = my_simdfsfdsgfs.run()

        with assert_things_are_printed(things=['my_simdfsfdsgfs', 'xxxxx\nyyyyy\n']):
            compare_experiment_records(rec)


def test_view_modes():

    with experiment_testing_context(new_experiment_lib=True):

        @experiment_function
        def my_simdfsffsdfsfs(a=1):

            print('xxxxx')
            print('yyyyy')
            return a+2

        my_simdfsffsdfsfs.add_variant(a=2)
        my_simdfsffsdfsfs.add_variant(a=3)

        my_simdfsffsdfsfs.run()

        for view_mode in ('full', 'results'):
            my_simdfsffsdfsfs.browse(view_mode=view_mode, command = 'q')


if __name__ == '__main__':
    test_experiments_function_additions()
    test_experiment_function_ui()
    test_invalid_arg_text()
    test_invalid_arg_text_when_object_arg()
    test_simple_experiment_show()
    test_view_modes()



from artemis.config import check_or_create_artemis_config, get_artemis_config_value
from artemis.fileman.config_files import get_config_sections


def validip(ip):
    return ip.count('.') == 3 and  all(0<=int(num)<256 for num in ip.rstrip().split('.'))


def get_configured_machines():
    """
    :return: A dict<machine_name: {'ip': ip_address, 'username': username}>
    """
    config_filename = check_or_create_artemis_config()
    sections = get_config_sections(config_filename)
    ip_sections = [s for s in sections if s.startswith('remote:')]
    machines = {}
    for sec in ip_sections:
        machine_name = sec[len('remote:'):]
        ip = get_artemis_config_value(section=sec, option='ip')
        username = get_artemis_config_value(section=sec, option='username')
        machines[machine_name] = {'ip': ip, 'username': username}
    return machines


EXAMPLE_MACHINE = """[remote:my_machine1]
ip=123.456.789.01
username=pluto
"""


def get_remote_machine_info(machine_name):
    machines = get_configured_machines()
    if len(machines)==0:
        raise Exception('No machines are specified!, You can add machines to ~/.artemisrc, as \n{}'.format(EXAMPLE_MACHINE))
    if machine_name not in machines:
        raise Exception("Could not find '{}' in the list of machines: {}.  You can add it to ~/.artemisrc".format(machine_name, machines.keys()))
    return machines[machine_name]



from artemis.fileman.file_getter import get_file
from scipy.ndimage.io import imread
from scipy.misc import imresize
import os
# from PIL import Image

__author__ = 'peter'

IMAGE_COLLECTION = {
    'wanderer': 'http://i.imgur.com/lR6wTNw.jpg',
    'dinnertime': 'http://i.imgur.com/nRczfmr.jpg',
    'limbo': 'https://upload.wikimedia.org/wikipedia/commons/thumb/3/3c/Follower_of_Jheronimus_Bosch_Christ_in_Limbo.jpg/1024px-Follower_of_Jheronimus_Bosch_Christ_in_Limbo.jpg',
    'manchester_newyear': 'http://i.imgur.com/UDEIMHp.jpg',
    'lenna': 'https://upload.wikimedia.org/wikipedia/en/2/24/Lenna.png',
    'starry_night': 'https://upload.wikimedia.org/wikipedia/commons/thumb/e/ea/Van_Gogh_-_Starry_Night_-_Google_Art_Project.jpg/800px-Van_Gogh_-_Starry_Night_-_Google_Art_Project.jpg',
    'nektarfront': 'https://upload.wikimedia.org/wikipedia/commons/0/00/Tuebingen_Neckarfront.jpg',
    'the_shipwreck_of_the_minotaur': 'https://upload.wikimedia.org/wikipedia/commons/2/2e/Shipwreck_turner.jpg',
    'scream': 'https://en.wikipedia.org/wiki/The_Scream#/media/File:The_Scream.jpg',
    'femme_nue_assise': 'https://upload.wikimedia.org/wikipedia/en/8/8f/Pablo_Picasso%2C_1909-10%2C_Figure_dans_un_Fauteuil_%28Seated_Nude%2C_Femme_nue_assise%29%2C_oil_on_canvas%2C_92.1_x_73_cm%2C_Tate_Modern%2C_London.jpg',
    'composition_vii': 'https://upload.wikimedia.org/wikipedia/commons/b/b4/Vassily_Kandinsky%2C_1913_-_Composition_7.jpg',
    'dog': 'http://www.dogchannel.com/images/articles/breed_profile_images/canaan-dog.jpg',
    'sea_turtle': 'https://en.wikipedia.org/wiki/Turtle#/media/File:TurtleOmaha.jpg',
    'heron': 'https://upload.wikimedia.org/wikipedia/commons/e/e6/Heron_tricol_01.JPG',
    }


def get_image(name, size = None):
    """
    Get an image by name.
    :param name: A string identifying an image from our dictionary.
    :return:
    """
    assert name in IMAGE_COLLECTION, "We don't have the image '%s' in the gallary" % (name, )
    _, ext = os.path.splitext(IMAGE_COLLECTION[name])
    relative_path = os.path.join('images', name)+ext
    filename = get_file(
        relative_name = relative_path,
        url = IMAGE_COLLECTION[name],
        )
    im_array = imread(filename)
    if im_array.ndim==2:
        im_array = im_array[:, :, None] + [0, 0, 0]
    if size is not None:
        im_array = imresize(im_array, get_new_size(im_array.shape[:2], new_size=size))
    return im_array


def get_new_size(cy_and_cx, new_size):
    cy, cx = cy_and_cx
    ny, nx = new_size or (None, None)
    aspect_ratio = float(cy) / cx
    if ny is None and nx is None:
        return cy
    elif nx is None:
        nx = int(ny * aspect_ratio)
    elif ny is None:
        ny = int(nx / aspect_ratio)
    return ny, nx


if __name__ == '__main__':
    from matplotlib import pyplot as plt

    im_name = 'manchester_newyear'
    im = get_image(im_name)
    plt.imshow(im)
    plt.title(im_name)
    plt.show()



import json
import os
import sys
from collections import OrderedDict
import numpy as np
import pip
from six.moves import input

from artemis.fileman.config_files import get_config_value
from artemis.config import get_artemis_config_value
from artemis.remote.utils import get_ssh_connection


def get_remote_installed_packages(ip_address):
    '''
    This method queries a remote python installation about the installed packages.
    All necessary information is extracted from ~/.artemisrc
    :param address: Ip address of remote server
    :return:
    '''
    python_executable = get_artemis_config_value(section=ip_address, option="python")
    function = "%s -c 'import pip; import json; print json.dumps({i.key: i.version  for i in pip.get_installed_distributions() })' "%python_executable

    ssh_conn = get_ssh_connection(ip_address)
    stdin , stdout, stderr = ssh_conn.exec_command(function)
    err = stderr.read()
    if err:
        msg="Quering %s python installation at %s sent a message on stderr. If you are confident that the error can be ignored, catch this RuntimeError" \
            "accordingly. The error is: %s"%(ip_address, python_executable, err)
        raise RuntimeError(msg)

    installed_packages = json.loads(stdout.read())
    ssh_conn.close()
    return installed_packages


def install_packages_on_remote_virtualenv(ip_address, packages):
    '''
    This function installs every package in packages on the remote virtual environment specified by the ip_address in ~/.artemisrc.
    In case the remote pip install -U command returns anything that is not "Successfully installed" or "Requirement already up-to-date" on stdout,
    the user is informed and an error that is not a SNIMissingWarning or InsecurePlatformWarning is printed.
    :param ip_address: ip_address, whose virtualenv is being modified
    :param packages: a dict {key:version} of package name and associated version
    :return:
    '''
    if len(packages) == 0:
        return
    print("installing/upgrading remote packages ...")
    python_path = get_artemis_config_value(ip_address,"python")
    activate_path = os.path.join(os.path.dirname(python_path),"activate") # TODO: Make this work without the user using virtualenv
    activate_command = "source %s"%activate_path
    ssh_conn = get_ssh_connection(ip_address)
    for key,version in packages.items():
        install_command = "pip install -U %s==%s" %(key,version)
        function = "; ".join([activate_command, install_command])
        stdin , stdout, stderr =ssh_conn.exec_command(function)
        out = stdout.read()
        if "Successfully installed" in out or "Requirement already up-to-date" in out:
            pass
        else:
            print(("Error in installing %s==%s:" % (key,version)))
            err = stderr.read()
            err = "\n".join([s for s in err.strip().split("\n") if "SNIMissingWarning" not in s and "InsecurePlatformWarning" not in s])
            print(err)
    ssh_conn.close()
    print("... Done")


def check_diff_local_remote_virtualenv(ip_address, auto_install=None, auto_upgrade=None, ignore_warnings=True):
    '''

    This method compares a remote virtual environment with the currently active virtual environment. Any local package that is not installed at
    the remote virtual environment or is installed in a different version at the remote virtual environment is identified. If auto_install is left at None,
    the user is prompted to specify which missing package he/she wants to install. If it is set to True, all missing packges are installed. If set to False,
    the missing packages are ignored. The same holds for auto_upgrade for different versions of installed packages. If some packages failed to install on
    the remote virtual environment, the user is asked to continue or exit the program unless 'ignore_warnings' is set to True, in which case the program
    continues even when some packages failed to install/upgrade.
    By setting any boolean value for auto_install and auto_upgrade, as well as setting ignore_warnings to True, no user interaction will be prompted.
    :param ip_address:
    :param auto_install:
    :param auto_upgrade:
    :param ignore_warnings:
    :return:
    '''
    print(("="*10 + " Checking remote virtualenv %s "%ip_address + "="*10))
    remote_packages = get_remote_installed_packages(ip_address)
    local_packages = {i.key: i.version  for i in pip.get_installed_distributions(include_editables=False)}
    missing_packages = OrderedDict()
    different_versions = OrderedDict()
    for (local_key, local_version) in local_packages.items():
        if local_key not in remote_packages.keys():
            missing_packages[local_key] = local_version
        elif local_version != remote_packages[local_key]:
            different_versions[local_key] = (local_version, remote_packages[local_key])

    # Install missing packages
    chosen_packages_to_install_and_update = None
    if len(missing_packages) > 0:
        missing_packages_string = ", ".join(missing_packages.keys())
        chosen_packages_to_install_and_update = {}
        if auto_install is not None:
            if auto_install:
                chosen_packages_to_install_and_update = missing_packages
            else:
                print(("Some locally installed packages are missing in the remote virtualenv at %s. You chose not to install them. "
                  "The missing packages are: \n %s " % (ip_address,missing_packages_string)))
        else:
            print("The following locally installed packages are missing in the virtualenv at %s:" % ip_address)
            print('\n'.join(['%s: %s (%s)' % (i, key, missing_packages[key]) for i, key in enumerate(missing_packages.keys())]))
            valid = False
            while not valid:
                ix = input("Please specify in the format '1, 3, 4' all packages you want to install. Can be empty or 'all' ")
                ix=ix.strip(",")
                numbers = ix.split(",")
                if numbers[0] == "all" and len(numbers) == 1:
                    chosen_packages_to_install_and_update = missing_packages
                    valid = True
                elif not numbers[0] and len(numbers) == 1:
                    valid = True
                else:
                    try:
                        numbers = [int(n) for n in numbers]
                    except:
                        print(("Please format input correctly. The input recieved was: \n%s"%ix))
                        continue
                    if not np.all([n >= 0 and n < len(missing_packages) for n in numbers]):
                        print(("Please make sure every number is valid. The input recieved was: \n%s"%ix))
                        continue
                    for i, key in enumerate(missing_packages.keys()):
                        if i in numbers:
                            chosen_packages_to_install_and_update[key] = missing_packages[key]
                    valid=True

    # Upgrade different version packages
    if len(different_versions) > 0:
        if auto_upgrade is not None:
            if auto_upgrade:
                for key, versions in different_versions.items():
                    chosen_packages_to_install_and_update[key] = versions[0]
            else:
                print(("Some locally installed packages are installed in a different version than on the remote virtualenv at %s. You chose not to upgrade them. "
                  "The different packages are:" % (ip_address)))
                for key, versions in different_versions.items():
                    print(("%s: local %s, reote %s") % (key, versions[0],versions[1]))
        else:
            print("The following locally installed packages are installed in a different version than on the virtualenv at %s:" % ip_address)
            print('\n'.join(['%s: %s (local: %s) => (remote: %s)' % (i, key, different_versions[key][0], different_versions[key][1]) for i, key in enumerate(different_versions.keys())]))
            valid = False
            while not valid:
                ix = input("Please specify in the format '1, 3, 4' all remote packages you want to upgrade (downgrade) to the local version. Can be empty or 'all' ")
                ix=ix.strip(",")
                numbers = ix.split(",")
                if numbers[0] == "all" and len(numbers) == 1:
                    for key, versions in different_versions.items():
                        chosen_packages_to_install_and_update[key] = versions[0]
                    valid = True
                elif not numbers[0] and len(numbers) == 1:
                    valid = True
                else:
                    try:
                        numbers = [int(n) for n in numbers]
                    except:
                        print(("Please format input correctly. The input recieved was: \n%s"%ix))
                        continue
                    if not np.all([n >= 0 and n < len(different_versions) for n in numbers]):
                        print(("Please make sure every number is valid. The input recieved was: \n%s"%ix))
                        continue
                    for i, key in enumerate(different_versions.keys()):
                        if i in numbers:
                            chosen_packages_to_install_and_update[key] = different_versions[key][0]
                    valid=True
    if chosen_packages_to_install_and_update is not None:
        if len(chosen_packages_to_install_and_update) == 0:
            print ("virtualenv up-to-date")
        else:
            install_packages_on_remote_virtualenv(ip_address, chosen_packages_to_install_and_update)
            if not ignore_warnings:
                remote_packages = get_remote_installed_packages(ip_address)
                missing_packages = OrderedDict()
                different_versions = OrderedDict()
                for (local_key, local_version) in local_packages.items():
                    if local_key not in remote_packages.keys():
                        missing_packages[local_key] = local_version
                    elif local_version != remote_packages[local_key]:
                        different_versions[local_key] = (local_version, remote_packages[local_key])
                if len(missing_packages)>0 or len(different_versions)>0:
                    valid = False
                    while not valid:
                        ix=input("The following packages could not be installed or upgraded: \n %s \nDo you want to continue? (y/n)"% (",".join(missing_packages.keys() + different_versions.keys())))
                        ix=ix.strip().lower()
                        if ix in ("y", "n"):
                            valid=True

                    if ix == "n":
                        print("Exiting...")
                        sys.exit(0)
                    else:
                        print("Ignoring discrepancies...")
    
    print(("="*10 + " Done remote virtualenv %s "%ip_address + "="*10))



import time


class ProgressIndicator(object):

    def __init__(self, expected_iterations=None, name=None, update_every = (2, 'seconds'), post_info_callback = None, show_total=False, just_use_last=False):

        self._expected_iterations = expected_iterations
        if isinstance(update_every, str):
            assert update_every.endswith('s'), 'Only support second-counds (eg "5s" for now)'
            update_every = (int(update_every[:-1]), 'seconds')
        elif isinstance(update_every, (float, int)):
            update_every = (update_every, 'iterations')

        update_interval, update_unit = update_every
        assert update_unit in ('seconds', 'percent', 'iterations')

        if update_unit == 'percent':
            update_unit = 'iterations'
            update_interval = update_interval/100.*expected_iterations
        self.name = name
        self._update_unit = update_unit
        self._update_interval = update_interval
        self._start_time = time.time()
        self._should_update = {
            'iterations': self._should_update_iter,
            'seconds': self._should_update_time,
            }[self._update_unit]

        self._i = 0
        self._last_update = -float('inf')
        self._post_info_callback = post_info_callback
        self.just_use_last = just_use_last
        self._last_time = self._start_time
        self._last_progress = 0
        self.show_total = show_total

    def __call__(self, iteration = None):
        self.print_update(iteration)

    def print_update(self, progress=None):
        self._current_time = time.time()
        elapsed = self._current_time - self._start_time
        if self._expected_iterations is None:
            print ('Progress{}: {:.1f}s Elapsed.  {} calls averaging {:.2g} calls/s'.format(
                '' if self.name is None else ' of '+self.name,
                elapsed,
                self._i+1,
                (self._i+1)/elapsed
                ))
        else:
            if progress is None:
                progress = self._i
            frac = float(progress)/(self._expected_iterations-1) if self._expected_iterations>1 else 1.
            if self._should_update() or progress == self._expected_iterations-1:
                elapsed = self._current_time - self._start_time
                if self.just_use_last is True:
                    remaining = (self._current_time - self._last_time)/(frac - self._last_progress) * (1-frac) if frac > 0 else float('NaN')
                else:
                    remaining = elapsed * (1 / frac - 1) if frac > 0 else float('NaN')
                elapsed = self._current_time - self._start_time
                self._last_update = progress if self._update_unit == 'iterations' else self._current_time
                print('Progress{}: {}%.  {:.1f}s Elapsed, {:.1f}s Remaining{}.  {}  {} calls averaging {:.2g} calls/s'.format(
                    '' if self.name is None else ' of '+self.name,
                    int(100*frac),
                    elapsed,
                    remaining,
                    ', {:.1f}s Total'.format(elapsed+remaining) if self.show_total else '',
                    self._post_info_callback() if self._post_info_callback is not None else '',
                    self._i+1,
                    (self._i+1)/elapsed
                    ))
        self._i += 1

        if self.just_use_last is True:
            self._last_time = self._current_time
            self._last_progress = frac

    def get_elapsed(self):
        return time.time() - self._start_time

    def get_iterations(self):
        return self._i

    def _should_update_time(self):
        return self._current_time-self._last_update > self._update_interval

    def _should_update_iter(self):
        return self._i - self._last_update > self._update_interval



from artemis.fileman.file_getter import get_file_in_archive, get_archive
from artemis.fileman.smart_io import smart_load_video
from artemis.general.should_be_builtins import bad_value
import os

__author__ = 'peter'


def load_ilsvrc_video(identifier, size = None, resize_mode='scale_crop', cut_edges=True, cut_edges_thresh=5):
    """
    Load a file from the ILSVRC Dataset.  The first time this is run, it will download an 8GB file, so be patient.

    Note: If you are using the same videos repeatedly, and applying resizing, you may want to call this function as:
        memoize_to_disk(load_ilsvrc_video)(identifier, size, ...)
    This will save you time on future runs.

    :param identifier: The file-name of the video, not including the extension.  Eg: 'ILSVRC2015_train_00249001'
    :param size:
    :param cut_edges:
    :param cut_edges_thresh:
    :return:
    """

    print ('Downloading ILSVER2015... this may take a while...')
    archive_folder_path = get_archive(
        relative_path='data/ILSVRC2015',
        url='http://vision.cs.unc.edu/ilsvrc2015/ILSVRC2015_VID_snippets_final.tar.gz'
        )
    print ('Done.')
    subpath = \
        'ILSVRC2015/Data/VID/snippets/test' if 'test' in identifier else \
        'ILSVRC2015/Data/VID/snippets/val' if 'val' in identifier else \
        'ILSVRC2015/Data/VID/snippets/train/ILSVRC2015_VID_train_0001/' if os.path.exists(os.path.join(archive_folder_path, 'ILSVRC2015/Data/VID/snippets/train/ILSVRC2015_VID_train_0001/', identifier + '.mp4')) else \
        'ILSVRC2015/Data/VID/snippets/train/ILSVRC2015_VID_train_0002/' if os.path.exists(os.path.join(archive_folder_path, 'ILSVRC2015/Data/VID/snippets/train/ILSVRC2015_VID_train_0002/', identifier + '.mp4')) else \
        'ILSVRC2015/Data/VID/snippets/train/ILSVRC2015_VID_train_0003/' if os.path.exists(os.path.join(archive_folder_path, 'ILSVRC2015/Data/VID/snippets/train/ILSVRC2015_VID_train_0003/', identifier + '.mp4')) else \
        bad_value(identifier, 'Could not find identifier: {}'.format(identifier, ))

    print('Loading %s' % (identifier, ))
    full_path = get_file_in_archive(
        relative_path='data/ILSVRC2015',
        subpath=os.path.join(subpath, identifier+'.mp4'),
        url='http://vision.cs.unc.edu/ilsvrc2015/ILSVRC2015_VID_snippets_final.tar.gz'
        )
    video = smart_load_video(full_path, size=size, cut_edges=cut_edges, resize_mode=resize_mode, cut_edges_thresh=cut_edges_thresh)
    print('Done.')
    return video


if __name__ == '__main__':
    import itertools
    from artemis.plotting.db_plotting import dbplot, hold_dbplots

    identifiers = ['ILSVRC2015_train_00033009', 'ILSVRC2015_train_00033010', 'ILSVRC2015_train_00763000', 'ILSVRC2015_test_00004002']
    videos = [load_ilsvrc_video(identifier, size=(224, 224), cut_edges=True) for identifier in identifiers]

    for i in itertools.count(0):
        with hold_dbplots():
            for identifier, vid in zip(identifiers, videos):
                dbplot(vid[i%len(vid)], identifier, title='%s: %s' % (identifier, i%len(vid)))



import numpy as np
from artemis.plotting.easy_plotting import ezplot
__author__ = 'peter'


class DataContainer(object):

    def __init__(self, im, line, struct, text, number):
        self._im = im
        self._line = line
        self._struct = struct
        self._text = text
        self._number = number


def test_easy_plot():

    thing = DataContainer(
        im =np.random.randn(30, 40),
        line = np.sin(np.arange(100)/10.),
        struct = {'video': np.random.randn(17, 20, 30)},
        text = 'adsagfdsf',
        number = 5
        )
    ezplot(thing, hang = False)


def test_plot_wmat():

    wmat = np.random.randn(7, 28, 28)
    ezplot(wmat, hang = False)


if __name__ == '__main__':
    test_plot_wmat()
    test_easy_plot()



import os
from collections import OrderedDict
from shutil import rmtree
from artemis.general.display import surround_with_header
from artemis.general.should_be_builtins import izip_equal
from six.moves import input


def crawl_directory(directory, ignore_hidden = True):
    """
    Given a directory, return a dict representing the tree of files under that directory.

    :param directory: A string representing a directory.
    :return: A dict<file_or_dir_name: content> where:
        file_or_dir_name is the name of the file within the parent directory.
        content is either
        - An absolute file path (for files) or
        - A dictionary containing the output of crawl_directory for a subdirectory.
    """
    contents = os.listdir(directory)
    if ignore_hidden:
        contents = [c for c in contents if not c.startswith('.')]
    this_dir = OrderedDict()
    for c in contents:
        full_path = os.path.join(directory, c)
        if os.path.isdir(full_path):
            this_dir[c] = crawl_directory(full_path)
        else:
            this_dir[c] = full_path
    return this_dir


class DirectoryCrawler(object):

    def __init__(self, directory, ignore_hidden=True):
        if isinstance(directory, (list, tuple)):
            directory = os.path.join(*directory)
        assert os.path.exists(directory), 'Directory "{}" does not exist.'.format(directory)
        self.directory = directory
        self._contents = None
        self.ignore_hidden = ignore_hidden

    def refresh(self):
        self._contents = None

    def listdir(self, refresh=False, sortby=None, end_dirs_with_slash=True):
        if isinstance(sortby, str) and sortby.startswith('-'):
            sortby = sortby[1:]
            reverse = True
        else:
            reverse = False
        if refresh or self._contents is None:
            self._contents = os.listdir(self.directory)
            if sortby=='mtime':
                self._contents = sorted(self._contents, key = lambda item: os.path.getmtime(self.get_path(item)))
            elif sortby=='name':
                self._contents = sorted(self._contents)
            elif sortby is not None:
                raise AssertionError('Invalid value for sortby: {}'.format(sortby))
            if reverse:
                self._contents = self._contents[::-1]
            if end_dirs_with_slash:
                self._contents = [c+os.sep if self.isdir(c) else c for c in self._contents]
            if self.ignore_hidden:
                self._contents = [item for item in self._contents if not item.startswith('.')]
        return self._contents

    def isdir(self, item):
        return os.path.isdir(self.get_path(item))

    def __iter__(self):
        for item in self.listdir():
            yield item

    def values(self):
        for item in self:
            yield self[item]

    def items(self):
        for item, full_path in izip_equal(self, self.values()):
            yield item, full_path

    def subdirs(self):
        for item in self:
            full_path = os.path.join(self.directory, item)
            if os.path.isdir(full_path):
                yield full_path

    def get_path(self, item):
        return os.path.join(self.directory, item)

    def __getitem__(self, item):
        full_path = os.path.normpath(os.path.join(self.directory, item))
        if os.path.isdir(full_path):
            return DirectoryCrawler(full_path, ignore_hidden=self.ignore_hidden)
        elif os.path.exists(full_path):
            return full_path
        else:
            raise Exception('No such file or directory: "{}"'.format(full_path))

    def __str__(self):
        return '{}({}, ignore_hidden={})'.format(self.__class__.__name__, self.directory, self.ignore_hidden)


class DirectoryCrawlerUI(object):
    """
    Commands for Directory Crowler UI:

    <Enter>         Refresh
    del 1-4          Remove listed files 1-4 (will be prompted for confirmation)
    del all          Remove all listed files.
    cd 4            Go into the 4th directory.
    cd ..           Go back 1 directory
    sortby name     Sort by name
    sortby mtime    Sort by last modified time
    sortby -mtime   Sort by last modified time in descending order
    h               Print help
    q               Quit
    """

    def __init__(self, directory, show_num_items=False, sortby='name'):
        self.dc = DirectoryCrawler(directory)
        self.show_num_subs = show_num_items
        self.sortby = sortby

    def _list_directory_contents(self):
        self._files = self.dc.listdir(refresh=True, sortby=self.sortby)
        if self.show_num_subs:
            return '\n'.join('{}: {} ({} items)'.format(i, fname, len(self.dc[fname].listdir())) for i, fname in enumerate(self._files))
        else:
            return '\n'.join('{}: {}'.format(i, fname) for i, fname in enumerate(self._files))

    def launch(self):
        commands = {
            'h': self.help,
            'del': self.delete,
            'cd': self.cd,
            'sortby': self.sortby
            }

        redraw = True
        while True:
            if redraw:
                print(surround_with_header('Contents of {}'.format(self.dc.directory), width=100))
                print(self._list_directory_contents())
            redraw = True
            command = input('Enter Command, or h for help >>')
            cmd_and_args = command.split(' ')
            cmd, args = cmd_and_args[0], cmd_and_args[1:]
            if cmd=='':
                continue
            elif cmd=='q':
                break
            elif cmd not in commands:
                print('Command not found: {}'.format(cmd))
                redraw=False
            else:
                commands[cmd](*args)

    def help(self):
        print(self.__doc__)
        input('Press Enter to continue >>')

    def sortby(self, arg):
        self.sortby = arg

    def _get_paths_from_range(self, user_range):
        if user_range=='all':
            files = self._files
        elif '-' in user_range:
            start, end = user_range.split('-')
            files = self._files[int(start): int(end)+1]
        else:
            files = [self._files[int(user_range)]]
        return [self.dc.get_path(filename) for filename in files]

    def delete(self, user_range):
        paths = self._get_paths_from_range(user_range)
        response = input('\n'.join(paths) + '\n Will be deleted.  Type "yes" to confirm>>')
        if response=='yes':
            for p in paths:
                rmtree(p)

    def cd(self, rel_path):
        try:
            subpath = self._files[int(rel_path)]
        except ValueError:
            subpath = rel_path
        self.dc = self.dc[subpath]


if __name__ == '__main__':
    from artemis.fileman.local_dir import get_artemis_data_path
    dcu = DirectoryCrawlerUI(get_artemis_data_path(), show_num_items=True, sortby='mtime')
    dcu.launch()



from collections import namedtuple, OrderedDict
from abc import abstractmethod
from artemis.general.nested_structures import flatten_struct
from artemis.plotting.drawing_plots import redraw_figure
from artemis.plotting.easy_plotting import plot_data_dict
from matplotlib import pyplot as plt

from artemis.plotting.matplotlib_backend import get_plot_from_data

__author__ = 'peter'


class BaseStream(object):
    def __init__(self, update_every = 1, figure = None):
        self._plots = None
        self._counter = -1
        self._update_every = update_every
        self._plot_keys = None
        self._fig = figure
        # TODO: Allow plots to be updated every iteration but only rendered every N'th iteration.  Important for streaming.

    def update(self, name = None):
        """
        Update all plots.  Note that when calling this function, we assume that you've updated the data returned by the
        callback.  When you call update multiple times without updating the data it can lead to incorrect results on
        plots with history (samples will be repeated)
        :param name: The name of the plot to update.  If not specified, updates all plots.
        """
        self._counter += 1
        if self._counter % self._update_every != 0:
            return

        data_dict = self._get_data_structure()  # dict : str -> IPlot

        if self._plots is None or set(data_dict.keys()) != self._plot_keys:
            # Note - this causes us to reset all plots (including ones with history
            # every time a new plot comes in, but that's ok for now).
            if self._fig is not None:
                self._fig.clf()
            else:
                self._fig = plt.figure()
            self._plots = self._get_plots_from_first_data(data_dict)
            self._plot_keys = set(self._plots.keys())
            plot_data_dict(data_dict, plots = self._plots, hang = False, figure = self._fig)
        else:
            if name is None:  # Update all plots
                for k, v in data_dict.items():
                    self._plots[k].update(v)
                    self._plots[k].plot()
            else:
                self._plots[name].update(data_dict[name])
                self._plots[name].plot()

        redraw_figure(self._fig)

    @abstractmethod
    def _get_data_structure(self):
        """
        :return a dict<s    tr: data> where data is some form of plottable data
        """

    @abstractmethod
    def _get_plots_from_first_data(self, first_data):
        """
        :return: a dict<str: IPlot> containing the plots corresponding to each element of the data.
        """


class LiveStream(BaseStream):
    """
    Lets you automatically generate live plots from some arbitrary data structure returned by a callback.
    """

    def __init__(self, callback, custom_handlers = {}, plot_mode = 'live', plot_types = {}, update_every=1,
                 figure = None, **plot_preference_kwargs):
        """
        :param callback: Some function that takes no arguments and returns some object.
        :param custom_handlers: A dict<type: function>.  If there's an object of one of the listed types
            returned from your callback, the function will take that object and return plot data from it.
        :param plot_mode: {'live', 'static', 'image'} - Determines what kind of plots to make for the data.
            See get_plot_from_data.
        :param update_every: Use this to only update the plot periodically - generally for speed.
        :param plot_preference_kwargs: Get passed down to get_plot_from_data
        """
        assert hasattr(callback, '__call__'), 'Your callback must be callable.'
        self._callback = callback
        self._custom_handlers=custom_handlers
        self._plot_mode = plot_mode
        self._plot_preference_kwargs = plot_preference_kwargs
        self._plot_types = plot_types
        BaseStream.__init__(self, update_every=update_every, figure = figure)

    def add_plot_type(self, key, plot_type):
        """
        You can pre-specify a plot to use for a given key.  The fact that you can add them dynamically is used in
        dbplot, where the data structure can change dynamically.
        """
        self._plot_types[key] = plot_type

    def _get_data_structure(self):
        struct = self._callback()
        assert struct is not None, 'Your plotting-data callback returned None.  Probably you forgot to include a return statememnt.'

        flat_struct = flatten_struct(struct, custom_handlers=self._custom_handlers, detect_duplicates=False)  # list<*tuple<str, data>>
        return OrderedDict(flat_struct)

    def _get_plots_from_first_data(self, first_data):
        return {k: get_plot_from_data(v, mode = self._plot_mode, **self._plot_preference_kwargs)
            if k not in self._plot_types else self._plot_types[k] for k, v in first_data.items()}


LivePlot = namedtuple('PlotBuilder', ['plot', 'cb'])


class LiveCanal(BaseStream):
    """
    Lets you make live plots by defining a dict of LivePlot objects, which contain the plot type and the data callback.
    LiveCanal gives you more control over your plots than LiveStream
    """

    def __init__(self, live_plots, **kwargs):
        """
        :param live_plots: A dict<str: (LivePlot OR function)>.  If the value is a LivePlot, you specify the type of
            plot to create.  Otherwise, you just specify a callback function, and the plot type is determined automatically
            based on the data.
        :param kwargs: Passed up to BaseStream
        """
        self._live_plots = live_plots
        self._callbacks = {k: lp.cb if isinstance(lp, LivePlot) else lp for k, lp in live_plots.items()}
        BaseStream.__init__(self, **kwargs)

    def _get_data_structure(self):
        return {k: cb() for k, cb in self._callbacks.items()}

    def _get_plots_from_first_data(self, first_data):
        # first_data = dict(first_data)
        return {k: pb.plot if isinstance(pb, LivePlot) else get_plot_from_data(first_data[k], mode = 'live') for k, pb in self._live_plots.items()}



from __future__ import print_function

import Queue
import argparse
import atexit
import pickle
import signal
import sys
import threading
import time
from datetime import datetime

import os
from matplotlib import pyplot as plt
import warnings
import matplotlib
warnings.filterwarnings("ignore",category=matplotlib.cbook.mplDeprecation)

#sys.path.extend([os.path.dirname(os.path.dirname(os.path.dirname(__file__)))])
from artemis.fileman.local_dir import get_artemis_data_path, format_filename
from artemis.plotting.matplotlib_backend import set_server_plotting
from artemis.plotting.db_plotting import dbplot, hold_dbplots, set_dbplot_figure_size
from artemis.plotting.saving_plots import save_figure
from artemis.remote.plotting.utils import _queue_get_all_no_wait, handle_socket_accepts
from artemis.remote.utils import get_socket, one_time_send_to


def send_port_if_running_and_join():
    port_file_path = get_artemis_data_path("tmp/plot_server/port.info", make_local_dir=True)
    if os.path.exists(port_file_path):
        with open(port_file_path, 'r') as f:
            port = pickle.load(f)
        print(port)
        print("Your dbplot call is attached to an existing plotting server. \nAll stdout and stderr of this existing plotting server "
              "is forwarded to the process that first created this plotting server. \nIn the future we might try to hijack this and provide you "
              "with these data streams")
        print("Use with care, this functionality might have unexpected side issues")
        try:
            while(True):
                time.sleep(20)
        except KeyboardInterrupt:
            print(" Redirected Server killed")
            sys.exit()
    else:
        with open(port_file_path,"w") as f:
            pass


def write_port_to_file(port):
    atexit.register(remove_port_file)
    port_file_path = get_artemis_data_path("tmp/plot_server/port.info", make_local_dir=True)
    if os.path.exists(port_file_path):
        # print("port.info file already exists. This might either mean that you are running another plotting server in the background and want to start a second one.\nIn this case ignore "
        #       "this message. Otherwise a previously run plotting server crashed without cleaning up afterwards. \nIn this case, please manually delete the file at {}".format(port_file_path),
        #       file = sys.stderr)
        # Keep for later development
        pass
    with open(port_file_path, 'wb') as f:
        pickle.dump(port,f)


def remove_port_file():
    port_file_path = get_artemis_data_path("tmp/plot_server/port.info", make_local_dir=True)
    if os.path.exists(port_file_path):
        os.remove(port_file_path)


def save_current_figure(path=""):
    fig = plt.gcf()
    file_name = format_filename(file_string = '%T', current_time = datetime.now())
    if path != "":
        save_path = os.path.join(path,"%s.pdf"%file_name)
    else:
        save_path = get_artemis_data_path('output/{file_name}.png'.format(file_name=file_name))
    save_figure(fig,path=save_path)


class GracefulKiller:
    kill_now = False
    def __init__(self):
        signal.signal(signal.SIGINT, self.exit_gracefully)

    def exit_gracefully(self, signum, frame):
        self.kill_now = True


def run_plotting_server(address, port, client_address, client_port):
    """
    Address and port to listen on.
    :param address:
    :param port:
    :return:
    """

    # Get the first available socket starting from portand communicate it with the client who started this server
    sock, port = get_socket(address=address, port=port)
    write_port_to_file(port)
    max_number_clients = 100
    max_plot_batch_size = 2000
    sock.listen(max_number_clients)
    one_time_send_to(address=client_address,port=client_port,message=str(port))

    # We want to save and rescue the current plot in case the plotting server receives a signal.SIGINT (2)
    killer = GracefulKiller()

    # The plotting server receives input in a queue and returns the plot_ids as a way of communicating that it has rendered the plot
    main_input_queue = Queue.Queue()
    return_queue = Queue.Queue()
    # Start accepting clients' communication requests
    t0 = threading.Thread(target=handle_socket_accepts,args=(sock, main_input_queue, return_queue, max_number_clients))
    t0.setDaemon(True)
    t0.start()


    # Received exp_dir on first db_plot_message?
    exp_dir_received = False
    # Now, we can accept plots in the main thread!

    set_dbplot_figure_size(9,10)
    while True:
        if killer.kill_now:
            sock.close() # will cause handle_socket_accepts thread to terminate
            # The server has received a signal.SIGINT (2), so we stop receiving plots and terminate
            break
        # Retrieve data points that might have come in in the mean-time:
        client_messages = _queue_get_all_no_wait(main_input_queue, max_plot_batch_size)
        # client_messages is a list of ClientMessage objects
        if len(client_messages) > 0:
            return_values = []
            with hold_dbplots():
                for client_msg in client_messages:  # For each ClientMessage object
                    # Take apart the received message, plot, and return the plot_id to the client who sent it
                    plot_message = pickle.loads(client_msg.dbplot_message)  # A DBPlotMessage object (see plotting_client.py)
                    plot_message.dbplot_args['draw_now'] = False

                    if not exp_dir_received:
                        if "exp_dir" == plot_message.dbplot_args["name"]:
                            atexit.register(save_current_figure,(plot_message.dbplot_args["data"]))
                            exp_dir_received = True
                            if len(client_messages) == 1:
                                continue
                            else:
                                continue
                    axis = dbplot(**plot_message.dbplot_args)
                    axis.ticklabel_format(style='sci', useOffset=False)
                    return_values.append((client_msg.client_address, plot_message.plot_id))
                if not exp_dir_received:
                    atexit.register(save_current_figure)
                    exp_dir_received = True
                plt.rcParams.update({'axes.titlesize': 'small', 'axes.labelsize': 'small'})
                plt.subplots_adjust(hspace=0.4,wspace=0.6)
            for client, plot_id in return_values:
                return_queue.put([client,plot_id])
        else:
            time.sleep(0.1)






if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--reuse', action="store_true", help='set if you want to merge plots on this server into one shared memory')
    parser.add_argument('--address',type=str,default="",help='This is the address of the  plotting client that started the server.')
    parser.add_argument('--port',type=int,default=-1,help='This is the port on which the plotting client that started the server listens for the port of the plotting server.')
    args = parser.parse_args()
    set_server_plotting(False)  # This causes dbplot to configure correctly
    if args.reuse is True:
        # TODO: has not been tested thoroughly yet. For example if you spawn two processes using the same server at the same time, there might be a conflict about who is the first who
        # actually sets up the server, and who joins the existing server
        send_port_if_running_and_join()

    run_plotting_server("0.0.0.0",7100, args.address, args.port) # We listen to the whole internet and start with port 7000



from collections import namedtuple
import itertools

from six import string_types

from artemis.general.numpy_helpers import fast_array
from artemis.general.should_be_builtins import bad_value, izip_equal
import numpy as np
from six.moves import xrange, zip
import time

__author__ = 'peter'


SINGLE_MINIBATCH_SIZE = 'single'
FULL_MINIBATCH_SIZE = 'full'


def minibatch_index_generator(n_samples, minibatch_size, n_epochs = 1, final_treatment = 'stop', slice_when_possible = True):
    """
    Generates the indices for minibatch-iteration.

    :param n_samples: Number of samples in the data you want to iterate through
    :param minibatch_size: Number of samples in the minibatch
    :param n_epochs: Number of epochs to iterate for
    :param final_treatment: How to terminate.  Options are:
        'stop': Stop when you can no longer get a complete minibatch
        'truncate': Produce a runt-minibatch at the end.
    :param slice_when_possible: Return slices, instead of indices, as long as the indexing does not wrap around.  This
        can be more efficient, since it avoids array copying, but you have to be careful not to modify your source array.
    :yield: IIndices that you can use to slice arrays for minibatch iteration.
    """

    if minibatch_size==SINGLE_MINIBATCH_SIZE:
        for i in (xrange(n_samples*n_epochs) if not np.isinf(n_epochs) else itertools.count(0)):
            yield i % n_samples
        return

    true_minibatch_size = n_samples if minibatch_size == FULL_MINIBATCH_SIZE else \
        minibatch_size if isinstance(minibatch_size, int) else \
        bad_value(minibatch_size)
    remaining_samples = int(n_epochs * n_samples) if not np.isinf(n_epochs) else np.inf

    base_indices = np.arange(minibatch_size)
    standard_indices = (lambda: slice(i, i+minibatch_size)) if slice_when_possible else (lambda: base_indices+i)
    i = 0
    while remaining_samples>0:
        next_i = i + true_minibatch_size
        if remaining_samples < minibatch_size:  # Final minibatch case
            if final_treatment == 'stop':
                break
            elif final_treatment == 'truncate':
                yield np.arange(i, i+remaining_samples) % n_samples
                break
            else:
                raise Exception('Unknown final treatment: %s' % final_treatment)
        elif next_i < n_samples:  # Standard case
            segment = standard_indices()
        else:  # Wraparound case
            segment = np.arange(i, next_i) % n_samples
            next_i = next_i % n_samples

        yield segment
        i = next_i
        remaining_samples -= minibatch_size


def checkpoint_minibatch_index_generator(n_samples, checkpoints, slice_when_possible = True):
    """
    Generates minibatch indices that fill the space between checkpoints.  This is useful, for instance, when you want to test
    at certain points, and your predictor internally iterates through the minibatch sample by sample between those.
    :param n_samples: Number of samples in the data you want to iterate through
    :param checkpoints: An array of indices at which the "checkpoints" happen.  Minibatches will be sliced up by these
        indices.
    :param slice_when_possible: Return slices, instead of indices, as long as the indexing does not wrap around.  This
        can be more efficient, since it avoids array copying, but you have to be careful not to modify your source array.
    :yield: Indices that you can use to slice arrays for minibatch iteration.
    """
    checkpoints = np.array(checkpoints, dtype = int)
    assert len(checkpoints) > 1 and checkpoints[0] >= 0 and np.all(np.diff(checkpoints) > 0)
    checkpoint_divs = list(zip(checkpoints[:-1], checkpoints[1:]))
    if checkpoints[0] > 0:
        checkpoint_divs.insert(0, (0, checkpoints[0]))
    for start, stop in checkpoint_divs:
        if start/n_samples == stop/n_samples:  # No wrap
            if slice_when_possible:
                yield slice(start % n_samples, stop % n_samples)
            else:
                yield np.arange(start % n_samples, stop % n_samples)
        else:
            yield np.arange(start, stop) % n_samples


def zip_minibatch_iterate(arrays, minibatch_size, n_epochs=1, final_treatment = 'truncate'):
    """
    Yields minibatches from each array in arrays in sequence.
    :param arrays: A collection of arrays, all of which must have the same shape[0]
    :param minibatch_size: The number of samples per minibatch
    :param n_epochs: The number of epochs to run for (or 'inf' for an infinite iterator)
    :yield: len(arrays) arrays, each of shape: (minibatch_size, )+arr.shape[1:]
    """
    assert isinstance(arrays, (list, tuple)), 'You need to provide an array or collection of arrays.'
    assert len(arrays)>0, 'Need at least one array'
    total_size = arrays[0].shape[0]
    if minibatch_size==FULL_MINIBATCH_SIZE:
        minibatch_size=total_size
    if n_epochs=='inf':
        n_epochs=float('inf')
    assert isinstance(n_epochs, (int, float))
    assert all(a.shape[0] == total_size for a in arrays), 'All arrays must have the same length!  Lengths are: %s' % ([len(arr) for arr in arrays])

    for ixs in minibatch_index_generator(n_samples=total_size, minibatch_size=minibatch_size, n_epochs=n_epochs, final_treatment=final_treatment):
        yield tuple(a[ixs] for a in arrays)


IterationInfo = namedtuple('IterationInfo', ['iteration', 'epoch', 'sample', 'time', 'test_now', 'done'])


def iteration_info(n_samples, minibatch_size, test_epochs = None, n_epochs = 5):
    """
    Create an iterator that keeps track of the state of minibatch iteration, and simplifies the scheduling of tests.
    You can izip this iterator with one that returns your data.

    :param n_samples: Number of samples in the dataset.
    :param minibatch_size: Size of minibatches
    :param test_epochs: Epochs on which you'd like to run tests.  You can also enter
        'every', which will test once-per-epoch,
        'always', which will test on every iteration
        'never', which will never test.
        ('every', 0.2), which will test at every 0.2 epochs (for example)
    :yield: IterationInfo objects which contain info about the state of iteration.
    """
    # next_text_point = 0 if test_epochs is not None and len(test_epochs)>0 else None
    start_time = time.time()
    n_samples = float(n_samples)
    if minibatch_size==FULL_MINIBATCH_SIZE:
        minibatch_size = n_samples
    elif minibatch_size == SINGLE_MINIBATCH_SIZE:
        minibatch_size = 1
    elif not isinstance(minibatch_size, int):
        raise Exception('Unexpected value for minibatch_size: {}'.format(minibatch_size))
    if isinstance(test_epochs, str):
        assert test_epochs in ('always', 'never', 'every'), "test_epochs={} is not valid".format(test_epochs)
    elif isinstance(test_epochs, tuple):
        assert len(test_epochs)==2, "If you pass in a tuple for test epochs, it should be in the form ('every', period).  Get {}".format(test_epochs)
        name, period = test_epochs
        assert period > 0, 'Period must be a positive number, not {}'.format(period)
        assert name == 'every'
    elif n_epochs is None and isinstance(test_epochs, (list, tuple, np.ndarray)):
        n_epochs = test_epochs[-1]

    last_epoch = -float('inf')
    for i in itertools.count(0):
        epoch = i*minibatch_size/n_samples
        test_now = (
                True if test_epochs=='always' else
                False if test_epochs=='never' else
                np.floor(epoch)>np.floor(last_epoch) if test_epochs == 'every' else
                bad_value(test_epochs)
            ) if isinstance(test_epochs, string_types) else \
            np.floor(epoch/period) > np.floor(last_epoch/period) if isinstance(test_epochs, tuple) else \
            False if test_epochs is None else \
            np.searchsorted(test_epochs, epoch, side='right') > np.searchsorted(test_epochs, last_epoch, side='right')
        info = IterationInfo(
            iteration = i,
            epoch = epoch,
            sample = i*minibatch_size,
            time = time.time()-start_time,
            test_now = test_now,
            done = epoch >= n_epochs if n_epochs is not None else False
            )
        yield info
        last_epoch = epoch


def zip_minibatch_iterate_info(arrays, minibatch_size, n_epochs=None, test_epochs = None):
    """
    Iterate through minibatches of arrays and yield info about the state of iteration though training.

    :param arrays:
    :param arrays: A collection of arrays, all of which must have the same shape[0]
    :param minibatch_size: The number of samples per minibatch
    :param n_epochs: The number of epochs to run for
    :param test_epochs: A list of epochs to test at.
    :return: (array_minibatches, info)
        arrays is a tuple of minibatches from arrays
        info is an IterationInfo object returning information about the state of iteration.
    """
    if n_epochs is None:
        assert isinstance(test_epochs, (list, tuple, np.ndarray)), "If you don't specify n_epochs, you need to specify an array of test epochs."
        n_epochs = test_epochs[-1]
    for arrays, info in zip(
            zip_minibatch_iterate(arrays, minibatch_size=minibatch_size, n_epochs='inf'),
            iteration_info(n_samples=arrays[0].shape[0], minibatch_size=minibatch_size, test_epochs=test_epochs, n_epochs=n_epochs)
            ):
        yield arrays, info
        if info.done:
            break


def minibatch_index_info_generator(n_samples, minibatch_size, n_epochs, test_epochs = None, slice_when_possible=False):
    for ixs, info in zip(
            minibatch_index_generator(n_samples=n_samples, minibatch_size=minibatch_size, n_epochs=n_epochs, slice_when_possible=slice_when_possible),
            iteration_info(n_samples=n_samples, minibatch_size=minibatch_size, test_epochs=test_epochs, n_epochs=n_epochs)
            ):
        yield ixs, info


def minibatch_iterate(data, minibatch_size, n_epochs=1):
    """
    Yields minibatches in sequence.
    :param data: A (n_samples, ...) data array
    :param minibatch_size: The number of samples per minibatch
    :param n_epochs: The number of epochs to run for
    :yield: (minibatch_size, ...) data arrays.
    """
    if minibatch_size == FULL_MINIBATCH_SIZE:
        minibatch_size = len(data)
    end = len(data)*n_epochs
    ixs = np.arange(minibatch_size)
    while ixs[0] < end:
        yield data[ixs % len(data)]
        ixs+=minibatch_size


def minibatch_iterate_info(data, minibatch_size, n_epochs, test_epochs = None):
    """
    Iterate through data and yield info about the iteration
    :param data: An (n_samples, ...) array of data
    :param minibatch_size: Size of the minibatches
    :param n_epochs: Number of epochs to iterate for
    :param test_epochs: A list of epochs to test at.
    :return:(minibatch, info)
        minibatch is a minibatch of data (minibatch_size, ...)
        info is an IterationInfo object returning information about the state of iteration.
    """
    for arrays, info in zip(
            minibatch_iterate(data, minibatch_size=minibatch_size, n_epochs=n_epochs),
            iteration_info(n_samples=data.shape[0], minibatch_size=minibatch_size, test_epochs=test_epochs)
            ):
        yield arrays, info


def minibatch_process(f, minibatch_size, mb_args = (), mb_kwargs = {}, fixed_kwargs={}):
    """
    Process inputs through a function f in minibatches, and stitch together the outputs.

    :param f: A function of the form y=f(x1, x2, ...)  where x1.shape[0] == x2.shape[0] == y.shape[0].
        We assume that outputs of f are equivariant to permuations along the 0th axis of inputs of f.
        i.e. each row is processed independently.
    :param minibatch_size: The size of the minibatches in which to feed the date through y
    :param mb_args: Arguments that will be fed through in minibatches
    :param mb_kwargs: Keyword-Arguments that will be fed through in minibatches
    :param fixed_kwargs: Keyword arguments that will remain fixed.
    :return: The output y
    """
    all_mb_args = list(mb_args) + list(mb_kwargs.values())
    assert len(all_mb_args)>0, 'Need some input.'
    assert callable(f), 'f must be a function'
    n_samples = len(mb_args[0])
    assert all(len(arg) == n_samples for arg in all_mb_args)
    mb_kwarg_list = mb_kwargs.items()
    fixed_kwarg_list = list(fixed_kwargs.items())
    index_generator = minibatch_index_generator(n_samples = n_samples, n_epochs=1, minibatch_size=minibatch_size, final_treatment='truncate')
    ix = next(index_generator)
    first_output = f(*(a[ix] for a in mb_args), **dict([(k, v[ix]) for k, v in mb_kwarg_list]+fixed_kwarg_list))
    if first_output is None:
        for ix in index_generator:
            f(*(a[ix] for a in mb_args), **dict([(k, v[ix]) for k, v in mb_kwarg_list]+fixed_kwarg_list))
    else:
        output_shape = first_output.shape if minibatch_size==SINGLE_MINIBATCH_SIZE else first_output.shape[1:]
        results = np.empty((n_samples, )+output_shape, dtype=first_output.dtype)
        results[:len(first_output)] = first_output
        for ix in index_generator:
            results[ix] = f(*(a[ix] for a in mb_args), **dict([(k, v[ix]) for k, v in mb_kwarg_list]+fixed_kwarg_list))
        return results


def generator_pool(generator_generator):
    for generator in generator_generator:
        yield generator


def batchify_generator(generator_generator, batch_size, receive_input=False, out_format ='array'):
    """
    Best understood by example:

    Suppose we want to get batches of frames from video data.  Where the batch[t][i] is the frame after batch[t-1][i].

    e.g. Suppose we have 7 videos.  In the following, each column represents a batch of data, and rows represent the
    index within a batch.

        -------vid-1---------|-------vid-5-------|--vid-7--
        -----------vid-2-----------|--------vid-6---------|
        -----vid-3-------|----------vid-4------------------

    generator_genererator yields 7 generators, corresponding to each of the movies.
    Each of those generators is a frame-generator, which produces the frames in a given video.
    Here, we generate frames from each movie, and start a new movies whenever an old one stops, until there are no
    new movies to start.

    :param generator_generator: An generator which generates generators
    :param batch_size: The size if the batch you want to yield
    :param receive_input: Expect a "send" to this generatoer AFTER it yields.  (see Python coroutines)
    :param out_format: 'array' or 'tuple_of_arrays' currently supported.
    :yield: An array consisting of batch_size of the outputs of the subgenerator, batched together.
    """
    assert receive_input in (False, 'post'), 'pre-receive not yet implemented'

    total = batch_size

    assert out_format in ('array', 'tuple_of_arrays')
    generators = [next(generator_generator) for _ in range(batch_size)]
    while True:
        items = []
        for i in range(batch_size):
            while True:
                try:
                    items.append(next(generators[i]))
                    break
                except StopIteration:
                    total+=1
                    generators[i] = next(generator_generator)  # This will rais StopIteration when we're out of generators

        if out_format=='array':
            output= np.array(items)
        else:
            output = tuple(np.array(x) for x in zip(*items))

        if not receive_input:
            yield output
        elif receive_input=='post':
            received_signal = (yield output)  # Assume in_format=='array'
            for gen, sig in izip_equal(generators, received_signal):
                gen.send(sig)
            yield None
        else:
            raise Exception(receive_input)


import pytest
from artemis.config import get_artemis_config_value

from artemis.remote.child_processes import PythonChildProcess
from artemis.remote.nanny import Nanny
from artemis.remote.utils import get_local_ips, get_remote_artemis_path
import os,sys
from contextlib import contextmanager
from StringIO import StringIO

ip_addresses = [get_artemis_config_value(section="tests", option="remote_server",default_generator=lambda: "127.0.0.1")]
# ip_addresses=["127.0.0.1"]
if ip_addresses[0] not in get_local_ips():
    ip_addresses.append("127.0.0.1")

def get_test_functions_path(ip_address):
    if ip_address in get_local_ips():
        return os.path.join(os.path.dirname(__file__), "remote_test_functions.py")
    else:
        remote_path = get_remote_artemis_path(ip_address)
        return os.path.join(remote_path, "remote/remote_test_functions.py")



@contextmanager
def captured_output():
    '''
    Source: https://stackoverflow.com/a/17981937/2068168
    :return:
    '''
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


def test_simple_process():
    for ip_address in ip_addresses:
        nanny = Nanny()
        command = "python %s --callback=%s"%(get_test_functions_path(ip_address),"success_function")
        pyc = PythonChildProcess(name="Process", ip_address=ip_address,command=command)
        nanny.register_child_process(pyc,)
        with captured_output() as (out,err):
            nanny.execute_all_child_processes()
        assert "%s: Success"%(pyc.get_name()) == out.getvalue().strip()

@pytest.mark.parametrize("N", [2, 10])
def test_several_simple_processes(N):
    for ip_address in ip_addresses:
        nanny = Nanny()
        command = "python %s --callback=%s"%(get_test_functions_path(ip_address),"success_function")
        for i in range(N):
            pyc = PythonChildProcess(name="Process%i"%i, ip_address=ip_address,command=command)
            nanny.register_child_process(pyc,)
        with captured_output() as (out,err):
            nanny.execute_all_child_processes()
        out_value = out.getvalue().strip()
        for pyc in nanny.managed_child_processes.values():
            assert "%s: Success"%(pyc.get_name()) in out_value

def test_process_termination():
    for ip_address in ip_addresses:
        nanny = Nanny()
        command = "python %s --callback=%s"%(get_test_functions_path(ip_address),"count_low")
        pyc = PythonChildProcess(name="Process1", ip_address=ip_address,command=command)
        nanny.register_child_process(pyc,)
        command = "python %s --callback=%s"%(get_test_functions_path(ip_address),"count_high")
        pyc = PythonChildProcess(name="Process2", ip_address=ip_address,command=command)
        nanny.register_child_process(pyc,)
        with captured_output() as (out,err):
            nanny.execute_all_child_processes(time_out=1)
        check_text = "Child Process Process2 at %s did not terminate 1 seconds after the first process in cluster terminated. Terminating now."%ip_address
        assert check_text in out.getvalue() or check_text in err.getvalue()

def test_output_monitor():
    for ip_address in ip_addresses:
        nanny = Nanny()
        command = "python %s --callback=%s"%(get_test_functions_path(ip_address),"short_sleep")
        pyc = PythonChildProcess(name="Process1", ip_address=ip_address,command=command)
        nanny.register_child_process(pyc,monitor_if_stuck_timeout=5)
        command = "python %s --callback=%s"%(get_test_functions_path(ip_address),"count_high")
        pyc = PythonChildProcess(name="Process2", ip_address=ip_address,command=command)
        nanny.register_child_process(pyc,monitor_if_stuck_timeout=3)
        with captured_output() as (out,err):
            nanny.execute_all_child_processes(time_out=1)
        check_text1 = "Timeout occurred after 0.1 min, process Process1 stuck"
        check_text = "Child Process Process2 at %s did not terminate 1 seconds after the first process in cluster terminated. Terminating now."%ip_address
        assert check_text in out.getvalue() or check_text in err.getvalue()
        assert check_text1 in out.getvalue() or check_text1 in err.getvalue()


def test_iter_print():
    for ip_address in ip_addresses:
        nanny = Nanny()
        command = ["python","-u", get_test_functions_path(ip_address), "--callback=iter_print"]
        pyc = PythonChildProcess(name="P1",ip_address=ip_address,command=command)
        nanny.register_child_process(pyc)
        with captured_output() as (out, err):
            nanny.execute_all_child_processes(time_out=1)
        if pyc.is_local():
            assert str(out.getvalue().strip()) == "\n".join(["P1: %i"%i for i in [0,2,4,6,8]])
            assert str(err.getvalue().strip()) == "\n".join(["P1: %i"%i for i in [1,3,5,7,9]])
        else:
            assert "\r\n".join(["P1: %i" % i for i in range(10)]) == str(out.getvalue().strip())


if __name__ == "__main__":
    test_simple_process()
    test_several_simple_processes(2)
    test_several_simple_processes(10)
    test_output_monitor()
    test_iter_print()



from abc import abstractmethod

__author__ = 'peter'
from copy import deepcopy
import pickle
"""
This module is made to help create "PrimativeSpecifiers".  A Primative Specifier is an object that can be turned
into a primative data structure, independent of all code.  You might want to do this, for instance, if you want to save
a trained model, but you want to be able to load it in any environment (if your code changes, or someone else wants to
load your model into their code.).
"""

_SAVEABLE_CLASSES = {}


class Serializable(object):

    @abstractmethod
    def to_spec(self):
        """
        :return: An object that can be used to reconstruct this class.
        """

    @classmethod
    @abstractmethod
    def from_spec(cls, spec):
        """
        :param spec: An object of the form returned by to_spec.
        :return: An instance of the class
        """


def _register_saveable_class(saveable_class):
    assert issubclass(saveable_class, PrimativeSpecifier)
    class_name = saveable_class.__name__
    if class_name in _SAVEABLE_CLASSES:
        assert saveable_class is _SAVEABLE_CLASSES[class_name], 'Another object has already registered as "%s"' % (saveable_class, )
    _SAVEABLE_CLASSES[class_name] = saveable_class


def load_primative(primative):
    assert '__class__' in primative, 'Saveable object must have a __class__ attribute'
    assert primative['__class__'] in _SAVEABLE_CLASSES, 'No class by the name of "%s" is registered as saveable' % (primative['__class__'])
    d = primative.copy()
    del d['__class__']
    obj = object.__new__(_SAVEABLE_CLASSES[primative['__class__']])
    obj.__dict__.update(d)
    return obj


class PrimativeSpecifier(object):

    def __new__(cls, *args, **kwargs):
        _register_saveable_class(cls)
        return object.__new__(cls)

    def to_primative(self):
        obj_dict = {'__class__': self.__class__.__name__}
        assert '__class__' not in self.__dict__, 'The key "__class__ is reserved, and should not be used by object specifiers.'
        obj_dict.update(self.__dict__)
        return obj_dict

    def clone(self):
        return load_primative(deepcopy(self.to_primative()))




import itertools
import time
import types

__author__ = 'peter'
import numpy as np


class CheckPointCounter(object):
    """
    Scenerio: You have a loop, and want to do something periodically within that loop, but not
    on every iteration.  Use this object to tell you if you want to do that thing or not.

    You give it a series of checkpoints, and each time it is called, it will return the number
    of checkpoints it has passed since it was last called.  If you only want to know IF it's passed
    a checkpoint, you can treat the return as a boolean.
    """

    def __init__(self, checkpoints):
        self._checkpoints = checkpoints
        self._index = 0

    def check(self, progress):
        """
        :param progress: Indicator of the progress - on the same scale of the checkpoints.
        :return: (points_passed, done), where:
            points_passed is the number of checkpoints passed since the last call to this method,
            done is a boolean indicating whether the last checkpoint has been passed.

        Note that done will only be True of points_passed>0
        """
        counter = 0
        done = False

        while True:
            if self._index == len(self._checkpoints):
                done = True
                break
            elif progress < self._checkpoints[self._index]:
                break
            else:
                counter += 1
                self._index += 1

        return counter, done


class Checkpoints(object):
    """
    An object where you specify checkpoints and return true every time one of those checkpoints is passed.
    """

    def __init__(self, checkpoint_generator, default_units = None, skip_first = False):
        """
        :param checkpoint_generator: Can be:
            A generator object returning checkpoints
            A list/tuple/array of checkpoints
            ('even', interval)
            ('exp', first, growth)
            None
        """

        if isinstance(checkpoint_generator, str):
            assert default_units in ('sec', None)
            assert checkpoint_generator.endswith('s')
            checkpoint_generator = int(checkpoint_generator[:-1])
            default_units = 'sec'
        elif default_units is None:
            default_units = 'iter'

        assert default_units in ('iter', 'sec')
        self.default_units = default_units
        if isinstance(checkpoint_generator, tuple):
            distribution = checkpoint_generator[0]
            if distribution == 'even':
                interval, = checkpoint_generator[1:]
                checkpoint_generator = (interval*i for i in itertools.count(1))
            elif distribution == 'exp':
                first, growth = checkpoint_generator[1:]
                checkpoint_generator = (first*i*(1+growth)**(i-1) for i in itertools.count(0))
            else:
                raise Exception("Can't make a checkpoint generator {}".format(checkpoint_generator))
        elif isinstance(checkpoint_generator, (list, tuple, np.ndarray)):
            checkpoint_generator = iter(checkpoint_generator)
        elif isinstance(checkpoint_generator, (int, float)):
            step = checkpoint_generator
            checkpoint_generator = (step*i for i in itertools.count(0))
        else:
            assert isinstance(checkpoint_generator, types.GeneratorType)

        if skip_first:
            next(checkpoint_generator)

        self.checkpoint_generator = checkpoint_generator
        self._next_checkpoint = float('inf') if checkpoint_generator is None else next(checkpoint_generator)
        self._counter = 0
        self._start_time = time.time()

    def __call__(self, t=None):
        if t is None:
            t = self._counter if self.default_units == 'iter' else time.time() - self._start_time
        self._counter += 1
        if t >= self._next_checkpoint:
            while t >= self._next_checkpoint:
                self._next_checkpoint = next(self.checkpoint_generator)
            return True
        else:
            return False

    def get_count(self):
        return self._counter


_COUNTERS_DICT = {}


def do_every(interval, counter_id=None, units = None):
    """
    Return true periodically.  Eg.

        # Plot every 100'th image.
        for im in images:
            if do_every(100):
                plot(current_image)

    :param interval: A number saying how often to return true.
    :param counter_id: ID used to uniquely identify this call to do_every.  (useful if you have more than
        one do_every call)
    :param units: 'iter' or 'sec'
    :return: True if checkpoint has just been passed, otherwise false.
    """
    if counter_id not in _COUNTERS_DICT:
        _COUNTERS_DICT[counter_id] = Checkpoints(checkpoint_generator=interval, default_units=units)
    return _COUNTERS_DICT[counter_id]()



import inspect
import logging
from contextlib import contextmanager
import time
from artemis.general.display import IndentPrint

logging.basicConfig()
_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.INFO)


@contextmanager
def debug_section(name):
    parent_frame = inspect.currentframe().f_back.f_back
    parent_loc = parent_frame.f_code.co_filename
    _LOGGER.info('Launching {name} @ File "{loc}", line {line}'.format(name=name, loc=parent_loc, line=parent_frame.f_lineno))
    start_time = time.time()
    with IndentPrint():
        yield
    end_time = time.time()
    _LOGGER.info('... Finished {} after {:.4g}s'.format(name, end_time-start_time))



import hashlib
import pickle
from collections import OrderedDict
import itertools
import numpy as np
from six import string_types, next

_ALREADY_SEEN_CODE = 'dbf056790fabd3c7b79c1ddab7b7ee49'.encode('utf-8')
_END_CODE = 'e0abd6b36d6e295b6c8859cdffc773df'.encode('utf-8')


def fixed_hash_eq(obj1, obj2):
    """
    Return True if the objects have equal fixed-hashes.  You can use this instead of the "==" operator when comparing
    two objects that don't contain all python primatives.  All objects contained must be either primatives, numpy arrays,
    or extend the FixedHashObject interface.

    :param obj1: An arbitrary object
    :param obj2: Another arbitrary project,
    :return: A boolean, indicating whether their hashes are equal.
    """
    return compute_fixed_hash(obj1)==compute_fixed_hash(obj2)


def compute_fixed_hash(obj, try_objects=False, _hasher = None, _memo = None, _count=None):
    """
    Given an object, return a hash that will always be the same (not just for the lifetime of the
    object, but for all future runs of the program too).
    :param obj: Some nested container of primitives
    :param try_objects: Try to break into objects
    :param _hasher: (for internal use - note that this is stateful, so calling this function with this argument changes
        the hasher object)
    :param _memo: (for internal use - to remember hashed objects and avoid infinite recursion)
    :param _count: (for internal use - to uniquely identify objects with circular references)
    :return: A 32-character hexidecimal (0-f) hash code for your object.
    """
    if _memo is None:
        _memo = {}
    elif not isinstance(obj, (np.ndarray, int, float, bool)+string_types) and id(obj) in _memo:
        _hasher.update(_ALREADY_SEEN_CODE)
        _hasher.update(str(_memo[id(obj)]).encode('utf-8'))
        _hasher.update(_END_CODE)
        return _memo[id(obj)]

    if _count is None:
        _count = itertools.count()
    _memo[id(obj)] =next(_count)

    if _hasher is None:
        _hasher = hashlib.md5()

    kwargs = dict(_hasher=_hasher, try_objects=try_objects, _memo=_memo, _count=_count)

    _hasher.update(obj.__class__.__name__.encode('utf-8'))
    if isinstance(obj, np.ndarray):
        _hasher.update(pickle.dumps(obj.dtype, protocol=2))
        _hasher.update(pickle.dumps(obj.shape, protocol=2))
        _hasher.update(obj.tostring())
    elif isinstance(obj, (int, float, bool)+string_types) or (obj is None) or (obj in (int, str, float, bool)):
        _hasher.update(pickle.dumps(obj, protocol=2))
    elif isinstance(obj, (list, tuple)):
        for el in obj:
            compute_fixed_hash(el, **kwargs)
    elif isinstance(obj, set):
        for el in sorted(obj):
            compute_fixed_hash(el, **kwargs)
    elif isinstance(obj, dict):
        keys = obj.keys() if isinstance(obj, OrderedDict) else sorted(obj.keys())
        for k in keys:
            compute_fixed_hash(k, **kwargs)
            compute_fixed_hash(obj[k], **kwargs)
    elif isinstance(obj, FixedHashObject):  # See below... allows you to make custom hashables
        compute_fixed_hash(obj.get_hash_description(), **kwargs)
    elif hasattr(obj, 'memo_hashable'):  # Deprecated, just here for back-compatibility
        compute_fixed_hash(obj.memo_hashable(), **kwargs)
    elif try_objects:
        keys = sorted(obj.__dict__.keys())
        for k in keys:
            compute_fixed_hash(k, **kwargs)
            compute_fixed_hash(obj.__dict__[k], **kwargs)
    else:
        # TODO: Consider whether to pickle by default.  Note that pickle strings are not necessairly the same for identical objects.
        raise NotImplementedError("Don't have a method for hashing this %s" % (obj, ))

    _hasher.update(_END_CODE)  # Necessary to distinguish ([a, b], c) from ([a, b, c])
    result = _hasher.hexdigest()
    _memo[id(obj)] = result
    return result


class FixedHashObject(object):
    """
    Implement this interface to create an object that you can create fixed hashes form.  You can then call
    compute_fixed_hash on this object.
    """

    def get_hash_description(self):
        """
        Override this method to return some primitive (python ond numpy only) data structure that represents your object.
        This will be used to create the fixed hash.  Be careful.  It's up to you to ensure that the data stryucture you
        return fully represents the state of your object.
        :return: Any data structure consisting only of Python and Numpy primatives.
        """
        raise NotImplementedError('You need to override this method to compute a fixed hash this object: {}'.format(self))
